--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mock_turtle_pkg
--
-- description: top level package with public types, definitions and components.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2014-2018
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;
use work.mt_mqueue_pkg.all;

package mock_turtle_pkg is

  constant c_MT_DEBUG_MESSAGE_FIFO_SIZE : integer := 512;

  type t_mt_timing_if is record
    link_up    : std_logic;
    dac_value  : std_logic_vector(23 downto 0);
    dac_wr     : std_logic;
    time_valid : std_logic;
    tai        : std_logic_vector(39 downto 0);
    cycles     : std_logic_vector(27 downto 0);
    aux_locked : std_logic_vector(7 downto 0);
  end record t_mt_timing_if;

  constant c_DUMMY_MT_TIMING : t_mt_timing_if := (
    link_up    => '0',
    dac_value  => (others => '0'),
    dac_wr     => '0',
    time_valid => '0',
    tai        => (others => '0'),
    cycles     => (others => '0'),
    aux_locked => (others => '0'));

  type t_int_array is array(integer range<>) of integer;

  subtype t_maxcpu_range is natural range 0 to 7;

  type t_mt_cpu_config is record
    -- CPU memory sizes, in bytes
    memsize    : natural;
    -- Per CPU message queue config.
    hmq_config : t_mt_mqueue_config;
    rmq_config : t_mt_mqueue_config;
  end record t_mt_cpu_config;

  type t_mt_cpu_config_array is array(t_maxcpu_range) of t_mt_cpu_config;

  type t_mt_config is record
    app_id          : std_logic_vector(31 downto 0);
    cpu_count       : natural range 1 to 8;
    cpu_config      : t_mt_cpu_config_array;
    -- shared memory size, in bytes
    shared_mem_size : integer range 256 to 65536;
  end record t_mt_config;

  constant c_DEFAULT_MT_CONFIG : t_mt_config :=
    (
      app_id          => x"115790de",
      cpu_count       => 2,
      cpu_config      => (others => (32768,
                                c_MT_DEFAULT_MQUEUE_CONFIG,
                                c_MT_DEFAULT_MQUEUE_CONFIG)),
      shared_mem_size => 8192
      );

  constant c_MOCK_TURTLE_SDB : t_sdb_device := (
    abi_class     => x"0000",              -- undocumented device
    abi_ver_major => x"01",
    abi_ver_minor => x"00",
    wbd_endian    => c_SDB_ENDIAN_BIG,
    wbd_width     => x"7",                 -- 8/16/32-bit port granularity
    sdb_component => (
      addr_first  => x"0000000000000000",
      addr_last   => x"000000000001ffff",
      product     => (
        vendor_id => x"000000000000CE42",  -- CERN
        device_id => x"000090de",
        version   => x"00000001",
        date      => x"20141201",
        name      => "Mock-Turtle-Core   ")));

  type t_mt_stream_sink_in_array2d is
    array(t_maxcpu_range) of t_mt_stream_sink_in_array(0 to 7);
  type t_mt_stream_sink_out_array2d is
    array(t_maxcpu_range) of t_mt_stream_sink_out_array(0 to 7);

  subtype t_mt_stream_source_in_array2d is t_mt_stream_sink_out_array2d;
  subtype t_mt_stream_source_out_array2d is t_mt_stream_sink_in_array2d;

  component mock_turtle_core is
    generic (
      g_CONFIG            : t_mt_config := c_DEFAULT_MT_CONFIG;
      g_WITH_RMQ          : boolean     := TRUE;
      g_SYSTEM_CLOCK_FREQ : integer     := 62500000;
      g_WITH_WHITE_RABBIT : boolean     := FALSE);
    port (
      clk_i           : in  std_logic;
      rst_n_i         : in  std_logic;
      sp_master_o     : out t_wishbone_master_out;
      sp_master_i     : in  t_wishbone_master_in                                  := c_DUMMY_WB_MASTER_IN;
      dp_master_o     : out t_wishbone_master_out_array(0 to g_CONFIG.cpu_count-1);
      dp_master_i     : in  t_wishbone_master_in_array(0 to g_CONFIG.cpu_count-1) := (others => c_DUMMY_WB_MASTER_IN);
      rmq_src_o       : out t_mt_stream_source_out_array2d;
      rmq_src_i       : in  t_mt_stream_source_in_array2d                         := (others => (others => c_MT_DUMMY_SOURCE_IN));
      rmq_snk_o       : out t_mt_stream_sink_out_array2d;
      rmq_snk_i       : in  t_mt_stream_sink_in_array2d                           := (others => (others => c_MT_DUMMY_SINK_IN));
      host_slave_i    : in  t_wishbone_slave_in;
      host_slave_o    : out t_wishbone_slave_out;
      clk_ref_i       : in  std_logic                                             := '0';
      tm_i            : in  t_mt_timing_if                                        := c_DUMMY_MT_TIMING;
      gpio_o          : out std_logic_vector(31 downto 0);
      gpio_i          : in  std_logic_vector(31 downto 0)                         := (others => '0');
      hmq_in_irq_o    : out std_logic;
      hmq_out_irq_o   : out std_logic;
      notify_irq_o    : out std_logic;
      console_irq_o   : out std_logic);
  end component mock_turtle_core;

end mock_turtle_pkg;
