--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mt_mqueue_remote
--
-- description: Remote MQ implementation. Exchanges messages between CPU CBs in
-- remote nodes.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2014-2018
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

use work.wishbone_pkg.all;
use work.mt_mqueue_pkg.all;

entity mt_mqueue_remote is
  generic (
    g_CONFIG : t_mt_mqueue_config := c_MT_DEFAULT_MQUEUE_CONFIG);
  port (
    clk_i        : in  std_logic;
    rst_n_i      : in  std_logic;
    slave_i      : in  t_wishbone_slave_in;
    slave_o      : out t_wishbone_slave_out;
    --  From OUT rmq
    src_o        : out t_mt_stream_source_out_array(0 to g_CONFIG.slot_count-1);
    src_i        : in  t_mt_stream_source_in_array(0 to g_CONFIG.slot_count-1);
    --  To IN rmq
    snk_o        : out t_mt_stream_sink_out_array(0 to g_CONFIG.slot_count-1);
    snk_i        : in  t_mt_stream_sink_in_array(0 to g_CONFIG.slot_count-1);
    rmq_status_o : out std_logic_vector(g_CONFIG.slot_count-1 downto 0));
end mt_mqueue_remote;

architecture arch of mt_mqueue_remote is
  signal cpu_incoming_in  : t_slot_bus_in_array(0 to g_config.slot_count-1);
  signal cpu_incoming_out : t_slot_bus_out_array(0 to g_config.slot_count-1);
  signal cpu_outgoing_in  : t_slot_bus_in_array(0 to g_config.slot_count-1);
  signal cpu_outgoing_out : t_slot_bus_out_array(0 to g_config.slot_count-1);

  signal ep_incoming_in  : t_slot_bus_in_array(0 to g_config.slot_count-1);
  signal ep_incoming_out : t_slot_bus_out_array(0 to g_config.slot_count-1);
  signal ep_outgoing_in  : t_slot_bus_in_array(0 to g_config.slot_count-1);
  signal ep_outgoing_out : t_slot_bus_out_array(0 to g_config.slot_count-1);

  signal incoming_stat : t_slot_status_out_array(0 to g_config.slot_count-1);
  signal outgoing_stat : t_slot_status_out_array(0 to g_config.slot_count-1);
begin  -- rtl

  U_SI_Wishbone_Slave : entity work.mt_mqueue_wishbone_slave
    generic map (
      g_WITH_GCR => TRUE,
      g_CONFIG   => g_CONFIG)
    port map (
      clk_i             => clk_i,
      rst_n_i           => rst_n_i,
      incoming_status_i => incoming_stat,
      outgoing_status_i => outgoing_stat,
      incoming_o        => cpu_incoming_in,
      incoming_i        => cpu_incoming_out,
      outgoing_o        => cpu_outgoing_in,
      outgoing_i        => cpu_outgoing_out,
      slave_i           => slave_i,
      slave_o           => slave_o);

  gen_slots : for i in 0 to g_CONFIG.slot_count-1 generate

    U_Out_SlotX : entity work.mt_mqueue_slot
      generic map (
        g_CONFIG  => g_config.slot_config(i))
      port map (
        clk_i   => clk_i,
        rst_n_i => rst_n_i,
        stat_o  => outgoing_stat(i),
        inb_i   => cpu_outgoing_in(i),
        inb_o   => cpu_outgoing_out(i),
        outb_i  => ep_outgoing_in(i),
        outb_o  => ep_outgoing_out(i));

    U_Out_Tx : entity work.mt_rmq_tx
      generic map (
        g_CONFIG  => g_config.slot_config(i))
      port map (
        clk_i       => clk_i,
        rst_n_i     => rst_n_i,
        inb_stat_i  => outgoing_stat(i),
        inb_i       => ep_outgoing_out(i),
        inb_o       => ep_outgoing_in(i),
        src_o       => src_o(i),
        src_i       => src_i(i));

    U_In_SlotX : entity work.mt_mqueue_slot
      generic map (
        g_CONFIG  => g_config.slot_config(i))
      port map (
        clk_i   => clk_i,
        rst_n_i => rst_n_i,
        stat_o  => incoming_stat(i),
        outb_i  => cpu_incoming_in(i),
        outb_o  => cpu_incoming_out(i),
        inb_i   => ep_incoming_in(i),
        inb_o   => ep_incoming_out(i));

    U_In_Rx : entity work.mt_rmq_rx
      generic map (
        g_CONFIG  => g_config.slot_config(i))
      port map (
        clk_i            => clk_i,
        rst_n_i          => rst_n_i,
        outb_stat_i      => incoming_stat(i),
        outb_i           => ep_incoming_out(i),
        outb_o           => ep_incoming_in(i),
        snk_i            => snk_i(i),
        snk_o            => snk_o(i));

    rmq_status_o (i) <= not incoming_stat(i).empty;
  end generate gen_slots;

end architecture arch;
