--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mt_mqueue_wishbone_slave
--
-- description: Wishbone slave for MQ's Slot/Global control registers.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2014-2018
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.mt_mqueue_pkg.all;

entity mt_mqueue_wishbone_slave is
  generic (
    g_WITH_GCR : boolean;
    g_CONFIG   : t_mt_mqueue_config);
  port (
    clk_i             : in  std_logic;
    rst_n_i           : in  std_logic;

    incoming_status_i : in  t_slot_status_out_array(g_CONFIG.slot_count-1 downto 0);
    outgoing_status_i : in  t_slot_status_out_array(g_CONFIG.slot_count-1 downto 0);

    incoming_o        : out t_slot_bus_in_array(g_CONFIG.slot_count-1 downto 0);
    incoming_i        : in  t_slot_bus_out_array(g_CONFIG.slot_count-1 downto 0);

    outgoing_o        : out t_slot_bus_in_array(g_CONFIG.slot_count-1 downto 0);
    outgoing_i        : in  t_slot_bus_out_array(g_CONFIG.slot_count-1 downto 0);

    slave_i           : in  t_wishbone_slave_in;
    slave_o           : out t_wishbone_slave_out;

    irq_config_o      : out t_mt_irq_config);
end mt_mqueue_wishbone_slave;

architecture arch of mt_mqueue_wishbone_slave is

  constant c_GCR_SLOT_COUNT        : std_logic_vector(3 downto 0) := x"0";
  constant c_GCR_SLOT_STATUS       : std_logic_vector(3 downto 0) := x"1";
  constant c_GCR_SLOT_IRQ_MASK     : std_logic_vector(3 downto 0) := x"2";
  constant c_GCR_SLOT_IRQ_COALESCE : std_logic_vector(3 downto 0) := x"3";

  signal slot_num, slot_num_d0         : std_logic_vector(2 downto 0);
  signal gcr_sel, gcr_sel_d0           : std_logic;
  signal in_area_sel, in_area_sel_d0   : std_logic;
  signal out_area_sel                  : std_logic;

  signal wb_write : std_logic;
  signal wb_read  : std_logic;

  signal gcr_rd_data : std_logic_vector(31 downto 0);

  signal irq_config : t_mt_irq_config;
  signal slot_status : std_logic_vector(31 downto 0) := (others => '0');

begin  -- arch

  slot_num     <= slave_i.adr(18 downto 16);
  in_area_sel  <= '1' when slave_i.adr(15 downto 14) = "01" else '0';
  out_area_sel <= '1' when slave_i.adr(15 downto 14) = "10" else '0';
  gcr_sel      <= '1' when slave_i.adr(15 downto 14) = "00" else '0';

  wb_write <= slave_i.cyc and slave_i.stb and slave_i.we;
  wb_read  <= slave_i.cyc and slave_i.stb and not slave_i.we;

  gen_incoming_slots : for i in 0 to g_CONFIG.slot_count-1 generate
    incoming_o(i) <=
      (sel   => f_to_std_logic (in_area_sel = '1' and unsigned(slot_num) = i),
       hdr_n => slave_i.adr(13),
       adr   => slave_i.adr(12 downto 0),
       dat   => slave_i.dat,
       we    => wb_write,
       wmask => slave_i.sel);

    slot_status(i) <= not incoming_status_i(i).empty;
  end generate gen_incoming_slots;

  gen_outgoing_slots : for i in 0 to g_CONFIG.slot_count-1 generate
    outgoing_o(i) <=
      (sel   => f_to_std_logic (out_area_sel = '1' and unsigned(slot_num) = i),
       hdr_n => slave_i.adr(13),
       adr   => slave_i.adr(12 downto 0),
       dat   => slave_i.dat,
       we    => wb_write,
       wmask => slave_i.sel);
    slot_status(i+16) <= not outgoing_status_i(i).full;
  end generate gen_outgoing_slots;

  p_delay : process(clk_i)
  begin
    if rising_edge(clk_i) then
      slot_num_d0     <= slot_num;
      in_area_sel_d0  <= in_area_sel;
      gcr_sel_d0      <= gcr_sel;
    end if;
  end process p_delay;

  gen_p_read_data : if g_CONFIG.slot_count > 0 generate
    p_read_data : process(incoming_i, outgoing_i, gcr_rd_data,
                          in_area_sel_d0, gcr_sel_d0, slot_num_d0)
    begin
      if gcr_sel_d0 = '1' then
        slave_o.dat <= gcr_rd_data;
      elsif in_area_sel_d0 = '1' then
        slave_o.dat <= incoming_i(to_integer(unsigned(slot_num_d0))).dat;
      else
        slave_o.dat <= outgoing_i(to_integer(unsigned(slot_num_d0))).dat;
      end if;
    end process p_read_data;
  end generate gen_p_read_data;

  gen_with_gcr : if g_WITH_GCR generate
    p_gcr_regs : process(clk_i)
    begin
      if rising_edge(clk_i) then
        if rst_n_i = '0' then
          irq_config.mask_in   <= (others => '0');
          irq_config.mask_out  <= (others => '0');
          irq_config.threshold <= (others => '0');
          irq_config.timeout   <= (others => '0');
        else
          gcr_rd_data <= (others => '0');

          if wb_write = '1' and gcr_sel = '1' then
            case slave_i.adr(5 downto 2) is
              when c_GCR_SLOT_IRQ_MASK =>
                irq_config.mask_out  <= slave_i.dat(23 downto 16);
                irq_config.mask_in   <= slave_i.dat(7  downto 0);
              when c_GCR_SLOT_IRQ_COALESCE =>
                irq_config.threshold <= slave_i.dat(7  downto 0);
                irq_config.timeout   <= slave_i.dat(23 downto 16);
              when others => null;
            end case;
          else
            case slave_i.adr(5 downto 2) is
              when c_GCR_SLOT_IRQ_MASK =>
                gcr_rd_data(23 downto 16) <= irq_config.mask_out;
                gcr_rd_data(7  downto 0)  <= irq_config.mask_in;

              when c_GCR_SLOT_IRQ_COALESCE =>
                gcr_rd_data(7  downto 0)  <= irq_config.threshold;
                gcr_rd_data(23 downto 16) <= irq_config.timeout;

              when c_GCR_SLOT_STATUS =>
                gcr_rd_data <= slot_status;

              when c_GCR_SLOT_COUNT =>
                gcr_rd_data(7 downto 0)  <=
                  std_logic_vector(to_unsigned(g_CONFIG.slot_count, 8));
              when others => null;

            end case;
          end if;
        end if;
      end if;
    end process p_gcr_regs;
  end generate gen_with_gcr;

  slave_o.stall <= '0';
  slave_o.err   <= '0';
  slave_o.rty   <= '0';

  p_ack : process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        slave_o.ack <= '0';
      else
        slave_o.ack <= wb_read or wb_write;
      end if;
    end if;
  end process p_ack;

  irq_config_o <= irq_config;

end architecture arch;
