--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mt_rmq_rx
--
-- description: Remote message rx adapter
--
--------------------------------------------------------------------------------
-- Copyright CERN 2014-2018
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mt_mqueue_pkg.all;

entity mt_rmq_rx is

  generic (
    g_CONFIG : t_mt_mqueue_slot_config);
  port (
    clk_i       : in std_logic;
    rst_n_i     : in std_logic;

    --  Input
    snk_o       : out t_mt_stream_sink_out;
    snk_i       : in  t_mt_stream_sink_in;

    --  Output
    outb_i      : in t_slot_bus_out;
    outb_o      : out t_slot_bus_in;

    --  Status from output.
    outb_stat_i : in t_slot_status_out);
end mt_rmq_rx;

architecture arch of mt_rmq_rx is

  type t_state_enum is (WAIT_SLOT, CLAIM, READY);

  type t_state is record
    state : t_state_enum;
    addr : unsigned(12 downto 2);
  end record;

  signal state, n_state : t_state;
begin
  p_fsm_next: process (rst_n_i, state, outb_stat_i, snk_i)
  begin
    if rst_n_i = '0' then
      n_state <= (state => WAIT_SLOT,
                  addr => (others => 'X'));
      snk_o <= (pkt_ready => '0',
                ready => '0');
      outb_o <= (sel => '0',
                 hdr_n => 'X',
                 adr => (others => 'X'),
                 dat => (others => 'X'),
                 we  => 'X',
                 wmask => "XXXX");
    else
      case state.state is
        when WAIT_SLOT =>
          if outb_stat_i.full = '0' then
            --  A packet can be received.  Claim it.
            snk_o <= (pkt_ready => '1',
                      ready => '1');
            outb_o <= (sel => '1',
                       hdr_n => '0',
                       adr => c_mqueue_addr_command(12 downto 0),
                       dat => (c_mqueue_command_claim => '1', others => '0'),
                       we  => '1',
                       wmask => "1111");
            n_state <= (state => CLAIM,
                        addr => unsigned(c_mqueue_addr_header(12 downto 2)));
          else
            --  Wait until a slot is ready.
            snk_o <= (pkt_ready => '0',
                      ready => '0');
            outb_o <= (sel => '0',
                       hdr_n => 'X',
                       adr => (others => 'X'),
                       dat => (others => 'X'),
                       we  => 'X',
                       wmask => "XXXX");
            n_state <= state;
          end if;
        when CLAIM =>
          snk_o <= (pkt_ready => '1',
                    ready => '1');
          if snk_i.valid = '1' then
            --  Data available, store it.
            outb_o <= (sel   => '1',
                       hdr_n => not snk_i.hdr,
                       adr   => std_logic_vector(state.addr & "00"),
                       dat   => snk_i.data,
                       we    => '1',
                       wmask => "1111");
            if snk_i.last = '1' then
              if snk_i.hdr = '0' then
                --  End of payload.
                n_state <= (state => READY,
                            addr => (others => 'X'));
              else
                --  End of header.
                n_state <=
                  (state => CLAIM,
                   addr => unsigned(c_mqueue_addr_payload(12 downto 2)));
              end if;
            else
                n_state <= (state => CLAIM,
                            addr => state.addr + 1);
            end if;
          else
            outb_o <= (sel => '0',
                       hdr_n => 'X',
                       adr => (others => 'X'),
                       dat => (others => 'X'),
                       we  => 'X',
                       wmask => "XXXX");
            n_state <= state;
          end if;
        when READY =>
          --  Make the slot ready.
          snk_o <= (pkt_ready => '0',
                    ready => '0');
          outb_o <= (sel   => '1',
                     hdr_n => '0',
                     adr   => c_mqueue_addr_command(12 downto 0),
                     dat   => (c_mqueue_command_ready => '1',
                               others => '0'),
                     we    => '1',
                     wmask => "1111");
          n_state <= (state => WAIT_SLOT,
                      addr => (others => 'X'));
      end case;
    end if;
  end process;

  p_fsm_reg: process (clk_i, rst_n_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        state <= (state => WAIT_SLOT,
                  addr => (others => 'X'));
      else
        state <= n_state;
      end if;
    end if;
  end process;
end arch;
