-------------------------------------------------------------------------------
-- Title      : Mock Turtle Node Core
-- Project    : Mock Turtle
-------------------------------------------------------------------------------
-- File       : mt_rmq_tx.vhd
-- Author     : Tomasz Włostowski
-- Company    : CERN BE-CO-HT
-- Created    : 2014-04-01
-- Last update: 2018-04-05
-- Platform   : FPGA-generic
-- Standard   : VHDL'93
-------------------------------------------------------------------------------
-- Description:
--
-- Single outgoing (MT->world) slot of the Remote Message Queue.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2014 CERN
--
-- This source file is free software; you can redistribute it
-- and/or modify it under the terms of the GNU Lesser General
-- Public License as published by the Free Software Foundation;
-- either version 2.1 of the License, or (at your option) any
-- later version.
--
-- This source is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-- PURPOSE.  See the GNU Lesser General Public License for more
-- details.
--
-- You should have received a copy of the GNU Lesser General
-- Public License along with this source; if not, download it
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.gencores_pkg.all;
use work.mt_mqueue_pkg.all;

entity mt_rmq_tx is

  generic (
    g_CONFIG : t_mt_mqueue_slot_config);
  port (
    clk_i   : in std_logic;
    rst_n_i : in std_logic;

    --  Input
    inb_i : in t_slot_bus_out;
    inb_o : out t_slot_bus_in;

    --  Status from input.
    inb_stat_i : in t_slot_status_out;

    --  Output
    src_o : out t_mt_stream_source_out;
    src_i : in  t_mt_stream_source_in
    );

end mt_rmq_tx;

architecture rtl of mt_rmq_tx is
  type t_state_enum is (WAIT_SLOT, CLAIM, SEND_HEADER, SEND_PAYLOAD, DISCARD);

  type t_state is record
    state : t_state_enum;
    addr : unsigned(12 downto 2);
    pkt_last_addr : unsigned(10 downto 0);
  end record;

  signal state, n_state : t_state;

  signal header_last, payload_last : std_logic;
begin

  header_last <= f_to_std_logic
     (state.addr(5 downto 2) = to_unsigned(2**g_CONFIG.header_bits - 1, 4));
  payload_last <= f_to_std_logic (state.addr = state.pkt_last_addr);

  p_fsm: process (rst_n_i, state, src_i, inb_i, inb_stat_i,
                  header_last, payload_last)
  begin
    if rst_n_i = '0' then
      inb_o <= (sel   => '0',
                hdr_n => 'X',
                adr   => (others => 'X'),
                dat   => (others => 'X'),
                we    => 'X',
                wmask => "XXXX");
      src_o <= (data  => (others => 'X'),
                hdr   => 'X',
                valid => '0',
                last  => 'X',
                error => 'X');
      n_state <= (state => WAIT_SLOT,
                  addr  => (others => 'X'),
                  pkt_last_addr => (others => 'X'));
    else
      n_state <= state;
      src_o <= (data  => (others => 'X'),
                hdr   => 'X',
                valid => '0',
                last  => 'X',
                error => 'X');

      case state.state is
        when WAIT_SLOT =>
          if src_i.pkt_ready = '1' and inb_stat_i.empty = '0' then
            --  A packet can be read and sent.  Claim it.
            inb_o <= (sel   => '1',
                      hdr_n => '0',
                      adr   => c_mqueue_addr_command(12 downto 0),
                      dat   => (c_mqueue_command_claim => '1',
                                others => '0'),
                      we    => '1',
                      wmask => "1111");

            n_state <= (state => CLAIM,
                        addr  => unsigned(c_mqueue_addr_header(12 downto 2)),
                        pkt_last_addr => (others => 'X'));
          else
            --  Do nothing, wait until the packet is ready.
            inb_o <= (sel   => '0',
                      hdr_n => 'X',
                      adr   => (others => 'X'),
                      dat   => (others => 'X'),
                      we    => 'X',
                      wmask => "XXXX");
          end if;
        when CLAIM =>
          --  Read first header word.
          inb_o <= (sel   => '1',
                    hdr_n => '0',
                    adr   => std_logic_vector(state.addr & "00"),
                    dat   => (others => 'X'),
                    we    => '0',
                    wmask => "XXXX");
          n_state <= (state => SEND_HEADER,
                      addr  => state.addr,
                      pkt_last_addr => (others => 'X'));
        when SEND_HEADER =>
          if state.addr = unsigned(c_mqueue_addr_header(12 downto 2)) then
            --  Save packet length.
            n_state.pkt_last_addr <= unsigned(inb_i.dat(10 downto 0)) - 1;
          else
            n_state.pkt_last_addr <= state.pkt_last_addr;
          end if;

          if src_i.ready = '1' then
            --  Forward header.
            src_o <= (data  => inb_i.dat,
                      hdr   => '1',
                      valid => '1',
                      last  => header_last,
                      error => '0');
            if header_last = '1' then
              --  Read first payload word.
              inb_o <= (sel   => '1',
                        hdr_n => '1',
                        adr   => c_mqueue_addr_payload(12 downto 0),
                        dat   => (others => 'X'),
                        we    => '0',
                        wmask => "XXXX");
              n_state.state <= SEND_PAYLOAD;
              n_state.addr  <= unsigned(c_mqueue_addr_payload(12 downto 2));
            else
              --  Read next header word.
              inb_o <= (sel   => '1',
                        hdr_n => '0',
                        adr   => std_logic_vector(state.addr + 1) & "00",
                        dat   => (others => 'X'),
                        we    => '0',
                        wmask => "XXXX");
              n_state.addr <= state.addr + 1;
            end if;
          else
            inb_o <= (sel   => '1',
                      hdr_n => '0',
                      adr   => std_logic_vector(state.addr & "00"),
                      dat   => (others => 'X'),
                      we    => '0',
                      wmask => "XXXX");
          end if;
        when SEND_PAYLOAD =>
          --  Read next payload
          inb_o <= (sel   => '1',
                    hdr_n => '1',
                    adr   => std_logic_vector(state.addr + 1) & "00",
                    dat   => (others => 'X'),
                    we    => '0',
                    wmask => "XXXX");

          if src_i.ready = '1' then
            --  Send payload word
            src_o <= (data  => inb_i.dat,
                      hdr   => '0',
                      valid => '1',
                      last  => payload_last,
                      error => '0');
            if payload_last = '1' then
              --  Packet was sent, discard it.
              inb_o <= (sel   => '1',
                        hdr_n => '0',
                        adr   => c_mqueue_addr_command(12 downto 0),
                        dat   => (c_mqueue_command_discard => '1',
                                  others => '0'),
                        we    => '1',
                        wmask => "1111");
              n_state <= (state => DISCARD,
                          addr  => (others => 'X'),
                          pkt_last_addr => (others => 'X'));
            else
              n_state <= (state => SEND_PAYLOAD,
                          addr => state.addr + 1,
                          pkt_last_addr => state.pkt_last_addr);
            end if;
          end if;
        when DISCARD =>
          --  Need to wait one cycle so that empty status  is updated.
          n_state <= (state => WAIT_SLOT,
                      addr  => (others => 'X'),
                      pkt_last_addr => (others => 'X'));
          inb_o <= (sel   => '0',
                    hdr_n => 'X',
                    adr   => (others => 'X'),
                    dat   => (others => 'X'),
                    we    => 'X',
                    wmask => "XXXX");
      end case;
    end if;
  end process;

  p_fsm_update: process (clk_i, rst_n_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        state <= (state => WAIT_SLOT,
                  addr  => (others => 'X'),
                  pkt_last_addr => (others => 'X'));
      else
        state <= n_state;
      end if;
    end if;
  end process;
end rtl;
