//------------------------------------------------------------------------------
// CERN BE-CO-HT
// Mock Turtle
// https://gitlab.cern.ch/coht/mockturtle
//------------------------------------------------------------------------------
//
// unit name:   MTCPUControl
//
// description: A SystemVerilog Class to provide an abstraction of the
// MockTurtle core's CSR, including functions such as CPU reset, firmware
// loading, debug UART monitoring, etc.
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __MT_CPU_CSR_INCLUDED
 `define __MT_CPU_CSR_INCLUDED

 `include "regs/mt_cpu_csr_regs.vh"

 `define SMEM_OP_DIRECT       0
 `define SMEM_OP_ADD          1
 `define SMEM_OP_SUB          2
 `define SMEM_OP_SET          3
 `define SMEM_OP_CLEAR        4
 `define SMEM_OP_FLIP         5
 `define SMEM_OP_TEST_AND_SET 6

class MTCPUControl;

   protected CBusAccessor bus;
   protected uint32_t base;
   protected uint32_t app_id;
   protected uint32_t core_count;
   protected uint32_t core_memsizes[$];
   protected string dbg_msg_queue[];

   function new ( CBusAccessor bus_, input uint32_t base_);
      base = base_;
      bus = bus_;
   endfunction // new

   function uint32_t getAppID();
      return app_id;
   endfunction // getAppID

   function uint32_t getCoreCount();
      return core_count;
   endfunction // getCoreCount

   function uint32_t getCoreMemorySize(int core);
      return core_memsizes[core];
   endfunction // getCoreMemorySize

   task writel ( uint32_t r, uint32_t v );
      bus.write ( base + r, v );
   endtask // _write

   task readl (  uint32_t r, ref uint32_t v );
      uint64_t tmp;
      bus.read (base + r, tmp );
      v= tmp;
   endtask // readl

   task init();
      int i;

      $display("Init CSR driver");
      readl(`ADDR_MT_CPU_CSR_APP_ID, app_id);
      readl(`ADDR_MT_CPU_CSR_CORE_COUNT, core_count);

      core_count &= 'hf;

      $display("App ID: 0x%x", app_id);
      $display("Core count: %0d", core_count);

      dbg_msg_queue = new[core_count];

      for(i = 0; i < core_count; i++)
	begin
	   uint32_t memsize;
	   writel(`ADDR_MT_CPU_CSR_CORE_SEL, i);
	   readl(`ADDR_MT_CPU_CSR_CORE_MEMSIZE, memsize);
	   $display("Core %1d: %3d kB private memory", i, memsize/1024);
	   core_memsizes.push_back( memsize );
	end
   endtask // init

   task reset_core(int core, int reset);
      uint32_t rstr;

      readl(`ADDR_MT_CPU_CSR_RESET, rstr);

      if(reset)
	rstr |= (1<<core);
      else
	rstr &= ~(1<<core);
      writel(`ADDR_MT_CPU_CSR_RESET, rstr);
   endtask // enable_cpu

   task debug_int_enable(int core, int enable);
      uint32_t imsk;

      readl(`ADDR_MT_CPU_CSR_DBG_IMSK, imsk);
      if(enable)
	imsk |= (1<<core);
      else
	imsk &= ~(1<<core);
      writel(`ADDR_MT_CPU_CSR_DBG_IMSK, imsk);
   endtask // debug_int_enable

   task load_firmware(int core, string filename, bit check);
      integer f = $fopen(filename, "r");
      int     i, addr = 0;

      // Stop in case of error.
      if (f == 0)
        $exit;

      reset_core(core, 1);

      writel(`ADDR_MT_CPU_CSR_CORE_SEL, core);

      while(!$feof(f))
	begin
	   uint32_t tmp, data;

	   tmp = $fread(data, f);

	   writel(`ADDR_MT_CPU_CSR_UADDR, addr);
	   writel(`ADDR_MT_CPU_CSR_UDATA, data);
	   addr ++;
	end // while (!$feof(f))

      if (check)
        begin
           void'($rewind(f));

           for(i = 0; i < addr; i++)
	     begin
	        uint32_t rv, data;
	        writel(`ADDR_MT_CPU_CSR_UADDR, i);
	        readl(`ADDR_MT_CPU_CSR_UDATA, rv);

	        void'($fread(data, f));
                assert (rv == data) else
	          $fatal("verification error at %x, got %x, expected %x\n",
		         i, rv, data);

	     end
        end
      $fclose(f);
   endtask // load_firmware

   task check_consoles();
      int i,j;
      uint32_t rval, cpu_mask;

      readl(`ADDR_MT_CPU_CSR_DBG_POLL , cpu_mask);
      if (cpu_mask == 0)
	return;

      for(i = 0; i < core_count; i++)
	begin
	   if(!(cpu_mask & (1 << i)))
	     continue;

	   writel(`ADDR_MT_CPU_CSR_CORE_SEL, i);
	   readl(`ADDR_MT_CPU_CSR_DBG_MSG, rval);

	   if(rval == 0 || rval == 13 || rval == 10)
	     begin
		$display("DBG MSG Core %0d: %s", i, dbg_msg_queue[i]);
                dbg_msg_queue[i] = "";
	     end
           else
             dbg_msg_queue[i] = {dbg_msg_queue[i], rval[7:0]};
	end
   endtask

   task check_c2h_interrupts();
      uint32_t val;
      readl(`ADDR_MT_CPU_CSR_INT, val);
      if (val != 0)
        $display("C2H interrupts: %x", val);
   endtask

   task update();
      check_consoles();
   endtask // update

   task set_smem_op(int op);
      writel(`ADDR_MT_CPU_CSR_SMEM_OP, op);
   endtask // set_smem_op

endclass

`endif //  `ifndef __MT_CPU_CSR_INCLUDED
