//------------------------------------------------------------------------------
// CERN BE-CO-HT
// Mock Turtle
// https://gitlab.cern.ch/coht/mockturtle
//------------------------------------------------------------------------------
//
// unit name:   main
//
// description: A SystemVerilog testbench for the supplied SVEC MT demo.
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`include "mock_turtle_driver.svh"
`include "vhd_wishbone_master.svh"

`timescale 1ns/1ps

module main;

   reg rst_n = 0;
   reg clk_125m = 0;

   always #4ns clk_125m <= ~clk_125m;

   initial begin
      repeat(20) @(posedge clk_125m);
      rst_n = 1;
   end

   // the Device Under Test
   svec_mt_demo #
     (
      .g_sim_bypass_vme(1)
      )
   DUT
     (
      .rst_a_n_i           (rst_n),
      .clk_125m_pllref_p_i (clk_125m),
      .clk_125m_pllref_n_i (~clk_125m),
      .fp_led_line_oen_o   (),
      .fp_led_line_o       (),
      .fp_led_column_o     (),
      .fp_gpio1_a2b_o      (),
      .fp_gpio2_a2b_o      (),
      .fp_gpio34_a2b_o     (),
      .fp_gpio_b           (),
      .sim_wb_i            (Host.out),
      .sim_wb_o            (Host.in),
      .VME_AS_n_i          (),
      .VME_RST_n_i         (),
      .VME_WRITE_n_i       (),
      .VME_AM_i            (),
      .VME_DS_n_i          (),
      .VME_GA_i            (),
      .VME_BERR_b          (),
      .VME_DTACK_n_b       (),
      .VME_RETRY_n_o       (),
      .VME_RETRY_OE_o      (),
      .VME_LWORD_n_b       (),
      .VME_ADDR_b          (),
      .VME_DATA_b          (),
      .VME_BBSY_n_i        (),
      .VME_IRQ_n_o         (),
      .VME_IACK_n_i        (),
      .VME_IACKIN_n_i      (),
      .VME_IACKOUT_n_o     (),
      .VME_DTACK_OE_b      (),
      .VME_DATA_DIR_b      (),
      .VME_DATA_OE_N_b     (),
      .VME_ADDR_DIR_b      (),
      .VME_ADDR_OE_N_b     ());


   IVHDWishboneMaster Host
     (
      .clk_i   (DUT.clk_sys),
      .rst_n_i (DUT.rst_n_sys));


   //  MT base (after VME sdb and vic).
   const uint64_t base = 'h2_0000;

   uint64_t count;

   task automatic inc_counter(ref CWishboneAccessor acc);
      uint64_t val;

      acc.read(base + 32'h1_0000, val);
      if (val != count)
        begin
           $display("incorrect count (%x, expect %x)", val, count);
           $stop;
        end
      count++;
      acc.write(base + 32'h1_0000, count);
   endtask

   initial begin

      CWishboneAccessor acc;
      MockTurtleDriver drv;

      @(posedge DUT.rst_n_sys);
      @(posedge DUT.clk_sys);

      acc = Host.get_accessor();
      acc.set_mode (PIPELINED);

      //  MT is mapped at 128kB on the VME wishbone port
      drv = new(acc, base);

      drv.init();

      $display("Display config rom");
      drv.cfg_rom_display();

      $display("Loading binary for cpu #0");
      drv.load_firmware(0, "../sw/verif/verif.bin", 1'b0);

      $display("Start cpu #0 at %t", $time);
      drv.reset_core(0, 0);

      // Enable all console interrupts
      drv.m_csr.writel(`ADDR_MT_CPU_CSR_DBG_IMSK, 8'hff);

      // Enable HMQ IN irqs.
      drv.m_hmq.gcr_write(`MQUEUE_GCR_IRQ_MASK, 8'hff);

      //  Enable VIC interrupts (unmask)
      acc.write(32'h2008, 8'hff);

      count = 0;

      forever begin
         uint32_t val;
         uint64_t risr;

         //  Read VIC RISR
         acc.read(32'h2004, risr);
         if (risr[2])
           drv.m_csr.check_consoles();

         // Check notify interrupt
         if (risr[3])
           begin
              drv.m_csr.readl(`ADDR_MT_CPU_CSR_INT, val);
              if (val != 0)
                begin
                   //  Read counter for cpu #0 (in smem)
                   inc_counter(acc);

                   $display("Notify interrupts: %x, count: %d", val, count);
                   drv.m_csr.writel(`ADDR_MT_CPU_CSR_INT, val);
                end
           end

         // Check HMQ IN
         if (risr[0])
           begin
              drv.m_hmq.gcr_read(`MQUEUE_GCR_SLOT_STATUS, val);
              $display("HMQ IN irq: %x", val);
              if ((val & 16'hff) != 0)
                begin
                   uint32_t stat;
                   uint32_t len;
                   uint32_t dat;
                   $display("HMQ GCR status: %x", val);
                   drv.m_hmq.incoming_read (`MQUEUE_SLOT_STATUS, stat );
                   $display(" SLOT status: %x", stat);
                   drv.m_hmq.incoming_read (`MQUEUE_SLOT_HEADER + 0, len);
                   $display(" SLOT header: %x", len);
                   len = 2;
                   drv.m_hmq.outgoing_write(`MQUEUE_SLOT_COMMAND,
                                            `MQUEUE_CMD_CLAIM );
                   drv.m_hmq.outgoing_write(`MQUEUE_SLOT_HEADER, len);

                   for (int i = 0; i < len; i++)
                     begin
                        drv.m_hmq.incoming_read(`MQUEUE_SLOT_DATA + 4*i, dat);
                        $display(" SLOT data[%d]: %x", i, dat);
                        drv.m_hmq.outgoing_write(`MQUEUE_SLOT_DATA + 4*i, ~dat);
                     end
                   inc_counter(acc);
                   drv.m_hmq.incoming_write(`MQUEUE_SLOT_COMMAND,
                                            `MQUEUE_CMD_DISCARD );
                   drv.m_hmq.outgoing_write(`MQUEUE_SLOT_COMMAND,
                                            `MQUEUE_CMD_READY );
                end
           end
	 #1us;
      end

   end

endmodule // main
