/*
 * This work is part of the White Rabbit Node Core project.
 *
 * Copyright (C) 2013-2014 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 */

/*.
 * White Rabbit Node Core
 *
 * mqueue.h: MQ register definitions (Host side)
 */

#ifndef __TRTL_MQUEUE_H
#define __TRTL_MQUEUE_H

// Global Control Registers base address, relative to BASE_HMQ (SLOT_COUNT, SLOT_STATUS, interrupt control, etc).
// Common for all incoming/outgoing slots in the queue
#define TRTL_MQ_BASE_GCR       (0x0)
// Incoming slot base address, relative to BASE_HMQ
#define TRTL_MQ_BASE_IN(slot)  (0x4000 + ((slot) << 16))
// Outgoung slot base address, relative to BASE_HMQ
#define TRTL_MQ_BASE_OUT(slot) (0x8000 + ((slot) << 16))

// MQ slot registers, relative to the base address of each slot: TRTL_MQ_BASE_IN(slot_no) or TRTL_MQ_BASE_OUT(slot_no)
#define TRTL_MQ_SLOT_COMMAND 0
// Status register
#define TRTL_MQ_SLOT_STATUS 4
// Start of header block
#define TRTL_MQ_SLOT_HEADER_START 0x1000
// Start of data block
#define TRTL_MQ_SLOT_DATA_START 0x2000

// Layout of TRTL_MQ_SLOT_COMMAND register:

// Claim: prepares a slot to send a message (w/o)
#define TRTL_MQ_CMD_CLAIM (1<<24)

// Purge: erases all messages from a slot (w/o)
#define TRTL_MQ_CMD_PURGE (1<<25)

// Ready: pushes the message to the queue. (w/o)
#define TRTL_MQ_CMD_READY (1<<26)

// Discard: removes last message from the queue, advancing to the next one (w/o)
#define TRTL_MQ_CMD_DISCARD (1<<27)

// Size of the message to be sent, in words (w/o). Must be written together with the
// READY command, e.g.:
// writel (TRTL_MQ_CMD_READY | 10, TRTL_MQ_SLOT_COMMAND);
#define TRTL_MQ_CMD_MSG_SIZE_MASK 0xff
#define TRTL_MQ_CMD_MSG_SIZE_SHIFT 0

// Layout of TRTL_MQ_SLOT_STATUS register:

// [0] Slot is full
#define TRTL_MQ_SLOT_STATUS_FULL (1<<0)
// [1] Slot is empty
#define TRTL_MQ_SLOT_STATUS_EMPTY (1<<1)

// [15:8] Number of occupied entries
#define TRTL_MQ_SLOT_STATUS_OCCUPIED_SHIFT 8
#define TRTL_MQ_SLOT_STATUS_OCCUPIED_MASK  0xff00

// [31:28] log2(number of words in the slot).
#define TRTL_MQ_SLOT_STATUS_LOG2_WIDTH_SHIFT 28
#define TRTL_MQ_SLOT_STATUS_LOG2_WIDTH_MASK  0xf0000000

// [7:2] log2(number of entries in the slot).
#define TRTL_MQ_SLOT_STATUS_LOG2_ENTRIES_SHIFT 2
#define TRTL_MQ_SLOT_STATUS_LOG2_ENTRIES_MASK  0xfc

//
// MQ Global Control Registers.Adresses relative to TRTL_MQ_BASE_GCR:
//
#define TRTL_MQ_GCR_INCOMING_STATUS_MASK (0x0000ffff)

// Number of slots in this implementation of HMQ
#define TRTL_MQ_GCR_SLOT_COUNT 0
// EMPTY bits of all slots in a single register (for polling/IRQ status)
#define TRTL_MQ_GCR_SLOT_STATUS 4
// Interrupt mask (Outgoing: EMPTY, Incoming: not EMPTY)
#define TRTL_MQ_GCR_IRQ_MASK 8
// IRQ Coalescing register (reserved for future use)
#define TRTL_MQ_GCR_IRQ_COALESCE 12

// Layout of SLOT_COUNT register

// [7:0] Number of slots
#define TRTL_MQ_GCR_SLOT_COUNT_N_SHIFT 0
#define TRTL_MQ_GCR_SLOT_COUNT_N_MASK 0xff

// Layout of SLOT_STATUS register

// [7:0] Incoming slots status. Each bit indicates a not EMPTY status of the corresponding incoming slot (data are available).
#define TRTL_MQ_GCR_SLOT_STATUS_IN_SHIFT 0
#define TRTL_MQ_GCR_SLOT_STATUS_IN_MASK 0xff

// [23:16] Outgoing slots status. Each bit indicates a NOT FULL status of the corresponding outgoing slot (data can be sent).
#define TRTL_MQ_GCR_SLOT_STATUS_OUT_SHIFT 16
#define TRTL_MQ_GCR_SLOT_STATUS_OUT_MASK 0xffff0000

// Layout of IRQ_MASK register

// [7:0] Incoming slots status interrupt mask. Each bit enables generation of interrupt on not EMPTY status of the corresponding incoming slot.
#define TRTL_MQ_GCR_IRQ_MASK_OUT_SHIFT 0
#define TRTL_MQ_GCR_IRQ_MASK_OUT_MASK 0xff

// [23:16] Outgoing slots status interrupt mask. Each bit enables generation of interrupt on NOT FULL status of the corresponding outgoing slot.
#define TRTL_MQ_GCR_IRQ_MASK_IN_SHIFT 16
#define TRTL_MQ_GCR_IRQ_MASK_IN_MASK 0x00ff0000

// Layout of IRQ_COALSESCE register
// The register is left for future IRQ coalescing support (if ever needed)

#endif
