/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>
#include <linux/uaccess.h>

#include <hw/mockturtle_cpu_csr.h>
#include "mockturtle-drv.h"


/**
 * Set the reset bit of the CPUs according to the mask
 */
void trtl_cpu_reset_set(struct trtl_dev *trtl, uint8_t mask)
{
	uint32_t reg_val;

	reg_val = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_RESET);
	reg_val |= (mask & 0xFF);
	trtl_iowrite(trtl, reg_val, trtl->base_csr + MT_CPU_CSR_REG_RESET);
}

/**
 * Clear the reset bit of the CPUs according to the mask
 */
static void trtl_cpu_reset_clr(struct trtl_dev *trtl, uint8_t mask)
{
	uint32_t reg_val;

	reg_val = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_RESET);
	reg_val &= (~mask & 0xFF);
	trtl_iowrite(trtl, reg_val, trtl->base_csr + MT_CPU_CSR_REG_RESET);
}

/**
 * It returns the CPU reset status
 */
static ssize_t trtl_show_reset(struct device *dev,
			       struct device_attribute *attr,
			       char *buf)
{
	struct trtl_cpu *cpu = to_trtl_cpu(dev);
	struct trtl_dev *trtl = to_trtl_dev(dev->parent);
	uint32_t reg_val;

	reg_val = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_RESET);

	return sprintf(buf, "%d\n", !!(reg_val & (1 << cpu->index)));
}

/**
 * It assert or de-assert the CPU reset line
 */
static ssize_t trtl_store_reset(struct device *dev,
				struct device_attribute *attr,
				const char *buf, size_t count)
{
	struct trtl_cpu *cpu = to_trtl_cpu(dev);
	struct trtl_dev *trtl = to_trtl_dev(dev->parent);
	long val;

	if (kstrtol(buf, 0, &val))
		return -EINVAL;

	if (val)
		trtl_cpu_reset_set(trtl, (1 << cpu->index));
	else
		trtl_cpu_reset_clr(trtl, (1 << cpu->index));

	return count;
}
DEVICE_ATTR(reset, (S_IRUGO | S_IWUSR), trtl_show_reset, trtl_store_reset);

static struct attribute *trtl_cpu_attr[] = {
	&dev_attr_reset.attr,
	NULL,
};

static const struct attribute_group trtl_cpu_group = {
	.attrs = trtl_cpu_attr,
};

const struct attribute_group *trtl_cpu_groups[] = {
	&trtl_cpu_group,
	NULL,
};


/**
 * It loads a given application into the CPU memory
 */
static int trtl_cpu_firmware_load(struct trtl_cpu *cpu, void *fw_buf,
				  size_t count, loff_t off)
{
	struct trtl_dev *trtl = to_trtl_dev(cpu->dev.parent);
	uint32_t *fw = fw_buf, word, word_rb, irq_dbg_old;
	int size, offset, i, cpu_memsize, err = 0;

	/*
	 * The Debug console uses the CORE select. In order to avoid an IRQ burst
	 * due to the loading of a new firmware, we temporary disable the IRQs.
	 * The Debug IRQ handler, can be called anyway right before we disable it
	 * but the spinlock will prevent any conflict on the core selection
	 */
	irq_dbg_old = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);
	trtl_iowrite(trtl, 0, trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);

	/* Select the CPU memory to write */
	spin_lock(&trtl->lock_cpu_sel);
	trtl_iowrite(trtl, cpu->index, trtl->base_csr + MT_CPU_CSR_REG_CORE_SEL);
	cpu_memsize = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_CORE_MEMSIZE );

	if (off + count > cpu_memsize) {
		dev_err(&cpu->dev,
			"Cannot load firmware: size limit %d byte\n",
			cpu_memsize);
		err = -ENOMEM;
		goto out;
	}

	/* Calculate code size in 32bit word*/
	size = (count + 3) / 4;
	offset = off / 4;

	/* Reset the CPU before overwrite its memory */
	trtl_cpu_reset_set(trtl, (1 << cpu->index));

	/* Clean CPU memory */
	for (i = offset; i < cpu_memsize / 1024; ++i) {
		trtl_iowrite(trtl, i, trtl->base_csr + MT_CPU_CSR_REG_UADDR);
		udelay(1);
		trtl_iowrite(trtl, 0, trtl->base_csr + MT_CPU_CSR_REG_UDATA);
		udelay(1);
	}

	/* Load the firmware */
	for (i = 0; i < size; ++i) {
		word = cpu_to_be32(fw[i]);
		trtl_iowrite(trtl, i + offset,
			   trtl->base_csr + MT_CPU_CSR_REG_UADDR);
		udelay(1);
		trtl_iowrite(trtl, word, trtl->base_csr + MT_CPU_CSR_REG_UDATA);
		udelay(1);
		word_rb = trtl_ioread(trtl,
				    trtl->base_csr + MT_CPU_CSR_REG_UDATA);
		udelay(1);
		if (word != word_rb) {
			dev_err(&cpu->dev,
				"failed to load firmware (byte %d | 0x%x != 0x%x)\n",
				i, word, word_rb);
			err = -EFAULT;
			goto out;
		}
	}
out:
	spin_unlock(&trtl->lock_cpu_sel);
	trtl_iowrite(trtl, irq_dbg_old,
		     trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);
	return err;
}

static int trtl_cpu_firmware_dump(struct trtl_cpu *cpu, void *fw_buf,
				  size_t count, loff_t off)
{
	struct trtl_dev *trtl = to_trtl_dev(cpu->dev.parent);
	uint32_t *fw = fw_buf, word;
	int size, offset, i, cpu_memsize, err = 0;


	/* Calculate code size in 32bit word*/
	size = (count + 3) / 4;
	offset = off / 4;

	/* Select the CPU memory to write */
	spin_lock(&trtl->lock_cpu_sel);
	trtl_iowrite(trtl, cpu->index, trtl->base_csr + MT_CPU_CSR_REG_CORE_SEL);

	cpu_memsize = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_CORE_MEMSIZE );

	if (off + count > cpu_memsize) {
		dev_err(&cpu->dev, "Cannot dump firmware: size limit %d byte\n",
			cpu_memsize);
		err = -ENOMEM;
		goto out;
	}

	/* Stop the CPU before dumping its memory */
	trtl_cpu_reset_set(trtl, (1 << cpu->index));

	/* Dump the firmware */
	for (i = 0; i < size; ++i) {
		trtl_iowrite(trtl, i + offset,
			   trtl->base_csr + MT_CPU_CSR_REG_UADDR);
		udelay(1);
		word = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_UDATA);
		udelay(1);
		fw[i] = be32_to_cpu(word);
	}

out:
	spin_unlock(&trtl->lock_cpu_sel);
	return err;
}


static int trtl_cpu_simple_open(struct inode *inode, struct file *file)
{
	int m = iminor(inode);

	file->private_data = to_trtl_cpu(minors[m]);

	return 0;
}

/**
 * It writes a given firmware into a CPU
 */
static ssize_t trtl_cpu_write(struct file *f, const char __user *buf,
			      size_t count, loff_t *offp)
{
	struct trtl_cpu *cpu = f->private_data;
	void *lbuf;
	int err;

	if (!count)
		return -EINVAL;

	lbuf = vmalloc(count);
	if (!lbuf)
		return -ENOMEM;
	if (copy_from_user(lbuf, buf, count)) {
		err = -EFAULT;
		goto out_cpy;
	}

	err = trtl_cpu_firmware_load(cpu, lbuf, count, *offp);
	if (err)
		goto out_load;

	*offp += count;

out_load:
out_cpy:
	vfree(lbuf);
	return err ? err : count;
}

/**
 * It reads the firmware from a CPU
 */
static ssize_t trtl_cpu_read(struct file *f, char __user *buf,
			     size_t count, loff_t *offp)
{
	struct trtl_cpu *cpu = f->private_data;
	void *lbuf;
	int err;

	lbuf = vmalloc(count);
	if (!lbuf)
		return -ENOMEM;

	err = trtl_cpu_firmware_dump(cpu, lbuf, count, *offp);
	if (err)
		goto out_dmp;

	if (copy_to_user(buf, lbuf, count)) {
		err = -EFAULT;
		goto out_cpy;
	}

	*offp += count;

out_cpy:
out_dmp:
	vfree(lbuf);
	return err ? err : count;
}

const struct file_operations trtl_cpu_fops = {
	.owner = THIS_MODULE,
	.open  = trtl_cpu_simple_open,
	.write  = trtl_cpu_write,
	.read = trtl_cpu_read,
	.llseek = generic_file_llseek,
};
