/*
 * Copyright (C) 2017 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <linux/kernel.h>
#include <linux/idr.h>
#include <linux/types.h>
#include <linux/bitmap.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/kallsyms.h>

#include <hw/mockturtle_cpu_csr.h>

#include "mockturtle-drv.h"

static DEFINE_IDA(trtl_tty_ida);

static inline void trtl_tty_insert_flip_char(struct trtl_cpu *cpu,
					     unsigned char ch, char flag)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0)
	tty_insert_flip_char(cpu->tty_port.tty, ch, flag);
#else
	tty_insert_flip_char(&cpu->tty_port, ch, flag);
#endif
}

static inline void trtl_tty_flip_buffer_push(struct trtl_cpu *cpu)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0)
	tty_flip_buffer_push(cpu->tty_port.tty);
#else
	tty_flip_buffer_push(&cpu->tty_port);
#endif
}

/**
 * trtl_tty_handler_getchar - Get Character From soft-CPU
 * It retrieves a character from the soft-CPU serial interface
 * and it adds the character to the TTY buffer
 * @cpu: the soft-CPU to use
 */
static void trtl_tty_handler_getchar(struct trtl_cpu *cpu)
{
	struct trtl_dev *trtl = to_trtl_dev(cpu->dev.parent);
	char c;

	/* Select the CPU and read data */
	spin_lock(&trtl->lock_cpu_sel);
	trtl_iowrite(trtl, cpu->index,
		     trtl->base_csr + MT_CPU_CSR_REG_CORE_SEL);
	c = trtl_ioread(trtl,
			trtl->base_csr + MT_CPU_CSR_REG_DBG_MSG);
	spin_unlock(&trtl->lock_cpu_sel);

	if (!cpu->tty_port.tty) /* Not open */
		return;

	trtl_tty_insert_flip_char(cpu, c , TTY_NORMAL);
}


static irqreturn_t trtl_tty_handler(int irq, void *arg)
{
	struct trtl_dev *trtl = arg;
	uint32_t status;
	int i, max_irq = 32, lock;

	status = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_DBG_POLL);
do_irq:
	i = -1;
	while (status && likely(++i < trtl->n_cpu)) {
		lock = spin_is_locked(&trtl->lock_cpu_sel);
		if (!(status & 0x1) || lock) {
			status >>= 1;
			continue;
		}
		/* Get a character from the soft-CPU */
		trtl_tty_handler_getchar(&trtl->cpu[i]);
	}

	/* Check again the interrupt status */
	status = trtl_ioread(trtl, trtl->base_csr + MT_CPU_CSR_REG_DBG_POLL);
	if (max_irq > 0 && status) {
		max_irq--;
		goto do_irq;
	}

	/* Push all the collected characters */
	for (i = 0; i < trtl->n_cpu; ++i) {
		if (!trtl->cpu[i].tty_port.tty) /* Not open */
			continue;

		trtl_tty_flip_buffer_push(&trtl->cpu[i]);
	}

	return IRQ_HANDLED;
}

static int trtl_tty_open(struct tty_struct *tty, struct file *file)
{
	struct trtl_cpu *cpu = to_trtl_cpu(tty->dev->parent);
	struct trtl_dev *trtl = to_trtl_dev(cpu->dev.parent);
	uint32_t irq_mask;
	int err;

	err = tty_port_open(&cpu->tty_port, tty, file);
	if (err)
		return err;

	/* Enable the interrupt on the Mock Turtle */
	irq_mask = trtl_ioread(trtl,
		    trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);
	irq_mask |= (1 << cpu->index);
	trtl_iowrite(trtl, irq_mask,
		     trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);

	return 0;
}

static void trtl_tty_close(struct tty_struct *tty, struct file *file)
{
	struct trtl_cpu *cpu = to_trtl_cpu(tty->dev->parent);
	struct trtl_dev *trtl = to_trtl_dev(cpu->dev.parent);
	uint32_t irq_mask;

	/* Disable the interrupt on the Mock Turtle */
	irq_mask = trtl_ioread(trtl,
		    trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);
	irq_mask &= ~(1 << cpu->index);
	trtl_iowrite(trtl, irq_mask,
		     trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);

	tty_port_close(&cpu->tty_port, tty, file);
}


const struct tty_operations trtl_tty_ops = {
	.open = trtl_tty_open,
	.close = trtl_tty_close,
};

static const struct tty_port_operations null_ops = { };

static int trtl_tty_port_init(struct trtl_cpu *cpu)
{
	cpu->tty_index = ida_simple_get(&trtl_tty_ida, 0, TRTL_MAX_CPU_MINORS,
					GFP_KERNEL);
	if (cpu->tty_index < 0)
		return -1;
	tty_port_init(&cpu->tty_port);
	cpu->tty_port.ops = &null_ops;

	cpu->tty_dev = tty_register_device(&trtl_tty_driver,
					   cpu->tty_index,
					   &cpu->dev);
	if (IS_ERR_OR_NULL(cpu->tty_dev)) {
		tty_port_put(&cpu->tty_port);
		return PTR_ERR(cpu->tty_dev);
	}

	return 0;
}


static void trtl_tty_port_exit(struct trtl_cpu *cpu)
{
	tty_unregister_device(&trtl_tty_driver, cpu->tty_index);

	ida_simple_remove(&trtl_tty_ida, cpu->tty_index);
}


void trtl_tty_remove(struct trtl_dev *trtl)
{
	struct platform_device *pdev = to_pdev_dev(trtl);
	int i;

	/* Disable the interrupts on the Mock Turtle */
	trtl_iowrite(trtl, 0x0, trtl->base_csr + MT_CPU_CSR_REG_DBG_IMSK);

	free_irq(platform_get_irq(pdev, TRTL_IRQ_DBG), trtl);

	for (i = 0; i < trtl->n_cpu; ++i)
		trtl_tty_port_exit(&trtl->cpu[i]);
}


int trtl_tty_probe(struct trtl_dev *trtl)
{
	struct platform_device *pdev = to_pdev_dev(trtl);
	struct resource *r;
	int i, err;

	for (i = 0; i < trtl->n_cpu; ++i) {
		err = trtl_tty_port_init(&trtl->cpu[i]);
		if (err)
			goto err_port;
	}

	r = platform_get_resource(pdev, IORESOURCE_IRQ, TRTL_IRQ_DBG);
	if (r) {
		/* Enable debug interrupts */
		err = request_irq(r->start, trtl_tty_handler, 0,
				  r->name, trtl);
		if (err) {
			dev_err(&trtl->dev,
				"Cannot request IRQ %lld - we'll not receive debug messages\n",
				r->start);
			goto err_irq;
		}
	}

	return 0;

err_irq:
err_port:
	while(--i)
		trtl_tty_port_exit(&trtl->cpu[i]);
	return err;
}
