/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * SPDX-License-Identifier: LGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libmockturtle.h>
#include <errno.h>

/**
 * It embeds the header into the message
 * @param[in] msg the trtl message
 * @param[in] hdr the header you want to embed into the message
 */
void trtl_message_header_set(struct trtl_msg *msg,
			     struct trtl_proto_header *hdr)
{
	memcpy(msg->data, hdr, sizeof(struct trtl_proto_header));
	msg->datalen = sizeof(struct trtl_proto_header) / 4;
}

/**
 * It retrieves the header from the message
 * @param[in] msg the trtl message
 * @param[out] hdr the header from the message
 */
void trtl_message_header_get(struct trtl_msg *msg,
			     struct trtl_proto_header *hdr)
{
	memcpy(hdr, msg->data, sizeof(struct trtl_proto_header));
}


/**
 * It packs a message to send to the HMQ. The function uses the header to get
 * the payload size. Rembember that the payload length unit is the 32bit word.
 * Remind also that the TRTL VHDL code, will convert a given 32bit word between
 * little endian and big endian
 * @param[out] msg raw message
 * @param[in] hdr message header
 * @param[in] payload data
 */
void trtl_message_pack(struct trtl_msg *msg,
		       struct trtl_proto_header *hdr,
		       void *payload)
{
	void *data = msg->data;

	trtl_message_header_set(msg, hdr);
	if (!payload)
		return;
	memcpy(data + sizeof(struct trtl_proto_header), payload,
	       hdr->len * 4);
	msg->datalen += hdr->len;
}


/**
 * it unpacks a message coming from the HMQ by separating the header from
 * the payload. You will find the payload size in the header.
 * Rembember that the payload length unit is the 32bit word.
 * Remind also that the TRTL VHDL code, will convert a given 32bit word between
 * little endian and big endian
 * @param[in] msg raw message
 * @param[out] hdr message header
 * @param[out] payload data
 */
void trtl_message_unpack(struct trtl_msg *msg,
			 struct trtl_proto_header *hdr,
			 void *payload)
{
	void *data = msg->data;

	trtl_message_header_get(msg, hdr);
	if (!payload)
		return;
	memcpy(payload, data + sizeof(struct trtl_proto_header),
	       hdr->len * 4);
}


/**
 * It appends a TLV structure to the given message
 * @param[in] msg
 * @param[in] hdr message header to use
 * @param[in] tlv element to append
 */
void trtl_message_buffer_push(struct trtl_msg *msg,
			      struct trtl_proto_header *hdr,
			      struct trtl_structure_tlv *tlv)
{
	unsigned int offset = sizeof(struct trtl_proto_header) / 4 + hdr->len;
	void *data;

	msg->data[offset++] = tlv->index;
	msg->data[offset++] = tlv->size;
	data = &msg->data[offset];

	memcpy(data, tlv->buf, tlv->size);

	hdr->len += 2 + (tlv->size / 4);

	trtl_message_header_set(msg, hdr);
	msg->datalen = hdr->len + sizeof(struct trtl_proto_header) / 4;
}

/**
 * A TLV record containing a structure will be take from the message head.
 * The function will update the message lenght in the header by removing
 * the size occupied by the last record.
 *
 * @TODO perhaps re think it because it consumes the hdr. It is better to
 * keep it safe
 * @param[in] msg raw message
 * @param[in] offset where start to pop buffers in 32bit words
 * @param[in] hdr message header
 * @param[out] tlv TLV record containing a structure
 */
off_t trtl_message_buffer_pop(struct trtl_msg *msg, off_t offset,
			      struct trtl_proto_header *hdr,
			      struct trtl_structure_tlv *tlv)
{
	if (offset > hdr->len) {
		errno = EINVAL;
		return -1;
	}

	tlv->index = msg->data[offset++];
	tlv->size = msg->data[offset++];
	if (tlv->size / 4 > hdr->len - 2) {/* -2 for intex and size */
		errno = EINVAL;
		return -1; /* TLV greater than what is declared in header */
	}
	memcpy(tlv->buf, &msg->data[offset], tlv->size);
	offset += (tlv->size / 4);

	return offset;
}

/**
 * Retrieve the current Real-Time Application version running. This is a
 * synchronous message.
 * @param[in] trtl device token
 * @param[out] version FPGA, RT and GIT version
 * @param[in] hmq_in hmq slot index where send the message
 * @param[in] hmq_out hmq slot index where you expect the answer
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int trtl_rt_version_get(struct trtl_dev *trtl, struct trtl_rt_version *version,
			unsigned int hmq_in, unsigned int hmq_out)
{
	struct trtl_proto_header hdr;
	struct trtl_hmq *hmq;
	struct trtl_msg msg;
	int err;

	memset(&msg, 0, sizeof(struct trtl_msg));
	memset(&hdr, 0, sizeof(struct trtl_proto_header));
	hdr.msg_id = TRTL_MSG_ID_VERS_REQ;
	hdr.slot_io = (hmq_in << 4) | hmq_out;
	hdr.flags = TRTL_HEAD_FLAG_SYNC | TRTL_HEAD_FLAG_RPC;
	hdr.len = 0;
	trtl_message_pack(&msg, &hdr, NULL);

	hmq = trtl_hmq_open(trtl, hmq_in, TRTL_HMQ_INCOMING);
	if (!hmq)
		return -1;

	/* Send the message and get answer */
	err = trtl_hmq_send_and_receive_sync(hmq, hmq_out, &msg,
					     trtl_default_timeout_ms);
	trtl_hmq_close(hmq);
	if (err <= 0)
		return -1;

	trtl_message_unpack(&msg, &hdr, version);
	if (hdr.msg_id != TRTL_MSG_ID_VERS_ANS) {
		errno = ETRTL_INVALID_MESSAGE;
		return -1;
	}

	return 0;
}


/**
 * It checks if a Real-Time application is running and answering to messages
 * @param[in] trtl device token
 * @param[in] hmq_in hmq slot index where send the message
 * @param[in] hmq_out hmq slot index where you expect the answer
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int trtl_rt_ping(struct trtl_dev *trtl,
		 unsigned int hmq_in, unsigned int hmq_out)
{
	struct trtl_proto_header hdr;
	struct trtl_hmq *hmq;
	struct trtl_msg msg;
	int err;

	memset(&hdr, 0, sizeof(struct trtl_proto_header));
	hdr.msg_id = TRTL_MSG_ID_PING;
	hdr.slot_io = (hmq_in << 4) | hmq_out;
	hdr.flags = TRTL_HEAD_FLAG_SYNC | TRTL_HEAD_FLAG_RPC;
	hdr.len = 0;
	trtl_message_pack(&msg, &hdr, NULL);

	hmq = trtl_hmq_open(trtl, hmq_in, TRTL_HMQ_INCOMING);
	if (!hmq)
		return -1;

	/* Send the message and get answer */
	err = trtl_hmq_send_and_receive_sync(hmq, hmq_out, &msg,
					     trtl_default_timeout_ms);
	if (err <= 0)
		return -1;
	trtl_hmq_close(hmq);
	trtl_message_unpack(&msg, &hdr, NULL);
	if (hdr.msg_id != TRTL_MSG_ID_ACK) {
		errno = ETRTL_INVALID_MESSAGE;
		return -1;
	}

	return 0;
}


/**
 * Real implementation to read/write variables
 */
static inline int trtl_rt_variable(struct trtl_dev *trtl,
				   struct trtl_proto_header *hdr,
				   uint32_t *variables,
				   unsigned int n_variables)
{
	struct trtl_msg msg;
	struct trtl_hmq *hmq;
	int err;

	hmq = trtl_hmq_open(trtl, (hdr->slot_io >> 4 & 0xF), TRTL_HMQ_INCOMING);
	if (!hmq)
		return -1;

	hdr->flags |= TRTL_HEAD_FLAG_RPC;

	memset(&msg, 0, sizeof(struct trtl_msg));

	/* Send asynchronous message, we do not wait for answers  */
	trtl_message_pack(&msg, hdr, variables);
	if (hdr->flags & TRTL_HEAD_FLAG_SYNC) {
		err = trtl_hmq_send_and_receive_sync(hmq, (hdr->slot_io & 0xF),
						     &msg, 1000);
		if (err > 0)
			trtl_message_unpack(&msg, hdr, variables);
		if (err == 0)
			err = -1;
	} else {
		err = trtl_hmq_send(hmq, &msg);
	}
	trtl_hmq_close(hmq);

	return err < 0 ? -1 : 0;
}


/**
 * It sends/receive a set of variables to/from the Real-Time application.
 *
 * The 'variables' field data format is the following
 *
 *     0     1     2     3     4     5    ...
 *  +-----+-----+-----+-----+-----+-----+
 *  | IDX | VAL | IDX | VAL | IDX | VAL | ...
 *  +-----+-----+-----+-----+-----+-----+
 *
 * IDX is the variable index defined by the real-time application
 * VAL is the associated value
 *
 * By setting the flag 'sync' you will send a synchronous message, otherwise
 * it is asyncrhonous. When synchronous the 'variables' field will be
 * overwritten by the syncrhonous answer; the answer contains the read back
 * values for the requested variable after the set operation. You can use
 * this to verify. You can use synchrounous messages to verify that you
 * variable are properly set.
 * This function will change the header content, in particular it will change
 * the following fields: msg_id, len
 * @param[in] trtl device token
 * @param[in|out] hdr header to send on input, header received on output
 * @param[in|out] var variables to get on input, variables values on output
 * @param[in] n_var number of variables to set. In other words,
 *            the number of indexes you have in the 'variables' fields.
 * @param[in] sync set if you want a synchronous message
 */
int trtl_rt_variable_set(struct trtl_dev *trtl,
			 struct trtl_proto_header *hdr,
			 uint32_t *variables,
			 unsigned int n_variables)
{
	int err;

	hdr->msg_id = TRTL_MSG_ID_VAR_SET;
	hdr->len = n_variables * 2;

	err = trtl_rt_variable(trtl, hdr, variables, n_variables);
	if (err)
		return err;
	if ((hdr->flags & TRTL_HEAD_FLAG_SYNC) &&
	    hdr->msg_id != TRTL_MSG_ID_VAR_GET_ANS) {
		errno = ETRTL_INVALID_MESSAGE;
		return -1;
	}

	return 0;
}


/**
 * It receive a set of variables from the Real-Time application.
 *
 * The 'variables' field data format is the following
 *
 *     0     1     2     3     4     5    ...
 *  +-----+-----+-----+-----+-----+-----+
 *  | IDX | VAL | IDX | VAL | IDX | VAL | ...
 *  +-----+-----+-----+-----+-----+-----+
 *
 * IDX is the variable index defined by the real-time application
 * VAL is the associated value
 *
 * This kind of message is always synchronous. The 'variables' field will be
 * overwritten by the syncrhonous answer; the answer contains the read back
 * values for the requested variables.
 * This function will change the header content, in particular it will change
 * the following fields: msg_id, flags, len
 * @param[in] trtl device token
 * @param[in|out] hdr header to send on input, header received on output
 * @param[in|out] var variables to get on input, variables values on output
 * @param[in] n_var number of variables to set. In other words,
 *            the number of indexes you have in the 'variables' fields
 */
int trtl_rt_variable_get(struct trtl_dev *trtl,
			 struct trtl_proto_header *hdr,
			 uint32_t *variables,
			 unsigned int n_variables)
{
	int err;

	hdr->msg_id = TRTL_MSG_ID_VAR_GET;
	/* Getting variables is always synchronous */
	hdr->flags |= TRTL_HEAD_FLAG_SYNC;
	hdr->len = n_variables * 2;

	err = trtl_rt_variable(trtl, hdr, variables, n_variables);
	if (err)
		return err;
	if (hdr->msg_id != TRTL_MSG_ID_VAR_GET_ANS) {
		errno = ETRTL_INVALID_MESSAGE;
		return -1;
	}

	return 0;
}


/**
 * Real implementation to read/write structures
 */
static int trtl_rt_buffer(struct trtl_dev *trtl,
			  struct trtl_proto_header *hdr,
			  struct trtl_structure_tlv *tlv,
			  unsigned int n_tlv)
{
	struct trtl_msg msg;
	struct trtl_hmq *hmq;
	int err, i;
	off_t offset = sizeof(struct trtl_proto_header) / 4;

	hmq = trtl_hmq_open(trtl, (hdr->slot_io >> 4 & 0xF), TRTL_HMQ_INCOMING);
	if (!hmq)
		return -1;

	hdr->flags |= TRTL_HEAD_FLAG_RPC;

	memset(&msg, 0, sizeof(struct trtl_msg));

	for (i = 0; i < n_tlv; ++i)
		trtl_message_buffer_push(&msg, hdr, &tlv[i]);

	if (hdr->flags & TRTL_HEAD_FLAG_SYNC) {
		err = trtl_hmq_send_and_receive_sync(hmq, (hdr->slot_io & 0xF),
						     &msg, 1000);
		if (err > 0) {
			for (i = 0; i < n_tlv; ++i) {
				offset = trtl_message_buffer_pop(&msg, offset,
								 hdr,
								 &tlv[i]);
				if (offset < 0) {
					err = -1;
					break;
				}
			}
		}
	} else {
		err = trtl_hmq_send(hmq, &msg);
	}
	trtl_hmq_close(hmq);

	return err <= 0 ? -1 : 0;
}


/**
 * It sends/receives a set of structures within TLV records.
 * This function will change the header content, in particular it will change
 * the following fields: msg_id, len
 * @param[in] trtl device token
 * @param[in|out] hdr header to send on input, header received on output
 * @param[in|out] tlv structures to get on input, structures values on output
 * @param[in] n_tlv number of tlv structures
 */
int trtl_rt_buffer_set(struct trtl_dev *trtl,
		       struct trtl_proto_header *hdr,
		       struct trtl_structure_tlv *tlv,
		       unsigned int n_tlv)
{
	int err;

	hdr->len = 0;
	hdr->msg_id = TRTL_MSG_ID_BUF_SET;

	err = trtl_rt_buffer(trtl, hdr, tlv, n_tlv);
	if (err)
		return err;
	if ((hdr->flags & TRTL_HEAD_FLAG_SYNC) &&
	    hdr->msg_id != TRTL_MSG_ID_BUF_GET_ANS) {
		errno = ETRTL_INVALID_MESSAGE;
		return -1;
	}

	return 0;
}

/**
 * It receives a set of structures within TLV records.
 * This function will change the header content, in particular it will change
 * the following fields: msg_id, flags, len
 * @param[in] trtl device token
 * @param[in|out] hdr header to send on input, header received on output
 * @param[in|out] tlv structures to get on input, structures values on output
 * @param[in] n_tlv number of tlv structures
 */
int trtl_rt_buffer_get(struct trtl_dev *trtl,
		       struct trtl_proto_header *hdr,
		       struct trtl_structure_tlv *tlv,
		       unsigned int n_tlv)
{
	int err;

	hdr->len = 0;
	hdr->msg_id = TRTL_MSG_ID_BUF_GET;
	/* Getting variables is always synchronous */
	hdr->flags |= TRTL_HEAD_FLAG_SYNC;

	err = trtl_rt_buffer(trtl, hdr, tlv, n_tlv);
	if (err)
		return err;
	if (hdr->msg_id != TRTL_MSG_ID_BUF_GET_ANS) {
		errno = ETRTL_INVALID_MESSAGE;
		return -1;
	}

	return 0;
}


/**
 * It prints the message header in a human readable format
 * @param[in] h message header
 */
void trtl_print_header(struct trtl_proto_header *h)
{
	fprintf(stdout,
		"    app_id 0x%x | msg_id %d | slot_io 0x%x | seq %d\n",
		h->rt_app_id, h->msg_id, h->slot_io, h->seq);
	fprintf(stdout,
		"    len %d | flags 0x%x | trans 0x%x | time %d\n",
		h->len, h->flags, h->trans, h->time);
}


/**
 * It prints the payload in a human readable format according
 * to the message type
 * @param[in] h message header
 * @param[in] buf message payload
 */
void trtl_print_payload(struct trtl_proto_header *h, void *buf)
{
	uint32_t *payload = buf;
	int i;

	switch (h->msg_id) {
	case TRTL_MSG_ID_VAR_SET:
	case TRTL_MSG_ID_VAR_GET:
	case TRTL_MSG_ID_VAR_GET_ANS:
		for (i = 0; i < h->len; i += 2) {
			fprintf(stdout, "[%d] 0x%08x\n",
				payload[i], payload[i + 1]);
		}
		break;
	default:
		for (i = 0; i < h->len; ++i) {
			if (i % 4 == 0)
				fprintf(stdout, "\n    %04d :", i);
			fprintf(stdout, " 0x%08x", payload[i]);
		}
		break;
	}
}


/**
 * It prints a message in a human readable format. This function assumes
 * that the message contains a Mock Turtle message header.
 * According to the message ID the format may change
 * @param[in] msg message to print
 */
void trtl_print_message(struct trtl_msg *msg)
{
	struct trtl_proto_header h;

	trtl_message_unpack(msg, &h, NULL);

	trtl_print_header(&h);
	if (h.len) {
		uint32_t payload[h.len];

		trtl_message_unpack(msg, &h, payload);
		trtl_print_payload(&h, payload);
	}
}
