/**
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include <errno.h>
#include <inttypes.h>
#include <mockturtle-framework.h>


/**
 * This is an @ref action_t function type. It fills the payload
 * witht an answer for the ping message.
 */
int rt_recv_ping(struct trtl_proto_header *hin, void *pin,
		 struct trtl_proto_header *hout, void *pout)
{
	rt_send_ack(hin, pin, hout, pout);
	return 0;
}


/**
 * This is an @ref action_t function type. It fills the payload with
 * version information.
 */
int rt_version_getter(struct trtl_proto_header *hin, void *pin,
		      struct trtl_proto_header *hout, void *pout)
{
	uint32_t *dout = pout;

	hout->msg_id = TRTL_MSG_ID_VERS_ANS;
	hout->len = sizeof(struct trtl_rt_version) / 4;
	rt_memcpy(dout, (uint32_t *)&_app->version,
		  sizeof(struct trtl_rt_version));

	return 0;
}


/**
 * This is an @ref action_t function type. Accorind the message request,
 * it overwrite a data structure with the one contained in the input
 * payload. If the message is syncrnous it will copy back the data structure.
 */
int rt_buffer_setter(struct trtl_proto_header *hin, void *pin,
			struct trtl_proto_header *hout, void *pout)
{
	unsigned int offset = 0, index, size;
	uint32_t *din = pin;

	if (!HAS_MOCKTURTLE_FRAMEWORK_BUFFER_ENABLE) {
		rt_send_nack(hin, pin, hout, NULL);
		return 0;
	}

	while (offset < hin->len) {
		pr_debug("%s: offset %d/%d\n\r", __func__, offset, hin->len);

		index = din[offset++];
		size = din[offset++];

		pr_debug("%s Type %d Len %d Addr 0x%p\n\r", __func__,
			  index, size, _app->buffers[index].buf);

		if (_app->buffers[index].len == size) {
			rt_memcpy((uint32_t *)_app->buffers[index].buf,
				  &din[offset], size);
		} else {
			pr_error("%s:%d structure %d len not correct %"PRId32" != %d\n\r",
				  __func__, __LINE__, index,
				  _app->buffers[index].len, size);
		}

		offset += (size / 4); /* Next TLV record */
	}

	/* Return back new values. Host can compare with what it sent
	   to spot errors */
	if (hin->flags & TRTL_HEAD_FLAG_SYNC)
		return rt_buffer_getter(hin, pin, hout, pout);

	return 0;
}

/**
 * This is an @ref action_t function type. Accorind the message request,
 * it copies one of the declared data structure to the output payload.
 */
int rt_buffer_getter(struct trtl_proto_header *hin, void *pin,
			struct trtl_proto_header *hout, void *pout)
{
	unsigned int offset = 0, index, size;
	uint32_t *din = pin;
	uint32_t *dout = pout;

	if (!HAS_MOCKTURTLE_FRAMEWORK_BUFFER_ENABLE) {
		rt_send_nack(hin, pin, hout, NULL);
		return 0;
	}

	hout->msg_id = TRTL_MSG_ID_BUF_GET_ANS;

	while (offset < hin->len) {
		pr_debug("%s: offset %d/%d\n\r", __func__, offset, hin->len);

		index = din[offset];
		dout[offset++] = index;
		size = din[offset];
		dout[offset++] = size;
		pr_debug("%s Type %d Len %d Addr 0x%p\n\r", __func__,
			  index, size, _app->buffers[index].buf);

		if (_app->buffers[index].len == size) {
			rt_memcpy(&dout[offset],
				  (uint32_t *)_app->buffers[index].buf,
				  size);
		} else {
			pr_error("%s: structure %d len not correct %"PRId32" != %d\n\r",
				  __func__, index, _app->buffers[index].len, size);
		}

		offset += (size / 4); /* Next TLV record */
	}

	return 0;
}

/**
 * This is an @ref action_t function type. Accorind the message request,
 * it writes a number of declared variables. If the message is synchronous
 * it copies back the values in the output payload.
 */
int rt_variable_setter(struct trtl_proto_header *hin, void *pin,
		       struct trtl_proto_header *hout, void *pout)
{
	struct rt_variable *var;
	uint32_t *din = pin, *mem, val;
	int i;

	if (!HAS_MOCKTURTLE_FRAMEWORK_VARIABLE_ENABLE) {
		pr_debug("Variable set action not supported\n\r");
		rt_send_nack(hin, pin, hout, NULL);
		return 0;
	}

	/* we always have a pair of values  */
	if (hin->len % 2)
		rt_send_nack(hin, pin, hout, pout);

	/* Write all values in the proper place */
	for (i = 0; i < hin->len; i += 2) {
		if (din[i] >= _app->n_variables)
			continue;
		var = &_app->variables[din[i]];
		mem = (uint32_t *) var->addr;
		val = ((din[i + 1] & var->mask) << var->offset);
		if (var->flags & TRTL_RT_VARIABLE_FLAG_FLD)
			*mem = (*mem & ~var->mask) | val;
		else
			*mem = val;

		pr_debug("%s index %"PRIu32"/%d | [0x%p] = 0x%08"PRIx32" <- 0x%08"PRIx32" (0x%08"PRIx32") | index in msg (%d/%d)\n\r",
			  __func__,
			  din[i], _app->n_variables - 1,
			  mem, *mem, val, din[i + 1],
			  i + 1, hin->len - 1);
	}

	/* Return back new values. Host can compare with what it sent
	   to spot errors */
	if (hin->flags & TRTL_HEAD_FLAG_SYNC)
		return rt_variable_getter(hin, pin, hout, pout);

	return 0;
}

/**
 * This is an @ref action_t function type. Accorind the message request,
 * it copies a number of declared variables.
 */
int rt_variable_getter(struct trtl_proto_header *hin, void *pin,
		       struct trtl_proto_header *hout, void *pout)
{
	struct rt_variable *var;
	uint32_t *dout = pout, *din = pin, *mem, val;
	int i;

	if (!HAS_MOCKTURTLE_FRAMEWORK_VARIABLE_ENABLE) {
		pr_debug("Variable get action not supported\n\r");
		rt_send_nack(hin, pin, hout, NULL);
		return 0;
	}

	if (!hout || !pout)
		return -1;

	/* we always have a pair of values  */
	if (hin->len % 2)
		return -1;

	hout->msg_id = TRTL_MSG_ID_VAR_GET_ANS;

	/* Write all values in the proper place */
	for (i = 0; i < hout->len; i += 2) {
		if (din[i] >= _app->n_variables) {
			dout[i] = ~0; /* Report invalid index */
			continue;
		}
		dout[i] = din[i];
		var = &_app->variables[dout[i]];
		mem = (uint32_t *) var->addr;
		val = (*mem >> var->offset) & var->mask;
		dout[i + 1] = val;

		pr_debug("%s index %"PRIu32"d/%d | [0x%p] = 0x%08"PRIx32" -> 0x%08"PRIx32" | index in msg (%d/%d)\n\r",
			  __func__,
			  dout[i], _app->n_variables - 1,
			  mem, *mem,  dout[i + 1],
			  i + 1, hin->len - 1);
	}

	return 0;
}

/**
 * List of standard actions
 */
static action_t *trtl_actions_in[] = {
	rt_recv_ping,
	rt_version_getter,
	rt_variable_setter,
	rt_variable_getter,
	rt_buffer_setter,
	rt_buffer_getter,
};


/**
 * It runs the action associated with the given identifier
 * @param[in] id action identifier
 * @param[in] msg input message for the action
 * @return 0 on success. A negative value on error
 */
static inline int rt_action_run(struct trtl_proto_header *hin, void *pin)
{
	action_t *action;
	struct trtl_msg out_buf;
	struct trtl_proto_header hout;
	void *pout;
	int err = 0;

	if (hin->msg_id >= __TRTL_MSG_ID_MAX) {
		pr_error("Invalid Message ID %d\n\r", hin->msg_id);
		return -EINVAL;
	}

	if (hin->msg_id < __TRTL_MSG_ID_MAX_USER) {
		/* User Actions */
		if (hin->msg_id >= _app->n_actions ||
		    !_app->actions[hin->msg_id]) {
			pr_error("Cannot dispatch User Message ID 0x%x\n\r",
				 hin->msg_id);
			return -EINVAL;
		}
		action = _app->actions[hin->msg_id];
	} else {
		/* Mock Turtle Standard Actions */
		if (ARRAY_SIZE(trtl_actions_in) < hin->msg_id - __TRTL_MSG_ID_MAX_USER) {
			pr_error("Cannot dispatch Mock Turtle Message ID 0x%x\n\r",
				 hin->msg_id);
			return -EINVAL;
		}
		action = trtl_actions_in[hin->msg_id - __TRTL_MSG_ID_MAX_USER];
	}

	pp_printf("Message Input\n\r");
	rt_print_header(hin);
	rt_print_data(pin, 8);


	if (!(hin->flags & TRTL_HEAD_FLAG_SYNC)) {
		/* Asynchronous message, then no output */
		return action(hin, pin, NULL, NULL);
	}

	/* Synchronous message */
	out_buf = rt_mq_claim_out(hin);
	/* Do not write directly the header on the buffer because it does not
	 work for fields size different than 32bit */
	pout = rt_proto_payload_get((void *) out_buf.data);
	rt_memcpy((uint32_t *)&hout, (uint32_t *)hin, sizeof(struct trtl_proto_header));

	err = action(hin, pin, &hout, pout);
	if (err)
		rt_send_nack(hin, pin, &hout, NULL);

	rt_proto_header_set((void *) out_buf.data, &hout);
	rt_mq_msg_send(&out_buf);

	pr_debug("Message Output\n\r");
	rt_print_header(&hout);
	rt_print_data(pout, 8);

	return err;
}



/**
 * It dispatch messages coming from a given message queue.
 * @param[in] mq_in MQ index within mq declaration in rt_application
 * @todo provide support for remote queue
 * @return 0 on success. -1 on error
 */
int rt_mq_action_dispatch(unsigned int mq_in)
{
	struct trtl_proto_header *header;
	unsigned int mq_in_slot = _app->mq[mq_in].index;
	uint32_t *msg;
	void *pin;
	int err = 0;

	/* HMQ control slot empty? */
	if (_app->mq[mq_in].type == TRTL_HMQ) {
		if (!(mq_poll() & ( 1 << mq_in_slot)))
			return -EAGAIN;
	} else {
		return -EAGAIN; /* Not used now */
	}

	/* Get the message from the HMQ */
	msg = mq_map_in_buffer(0, mq_in_slot);

	pr_debug("Incoming message\n\r");
	rt_print_data(msg, 8);

	header = rt_proto_header_get((void *) msg);
	rt_print_header(header);
	if (header->rt_app_id && header->rt_app_id != _app->version.rt_id) {
		pr_error("Not for this application 0x%x\n\r", header->rt_app_id);
		err = -EINVAL;
		goto out;
	}
	if (!(header->flags & TRTL_HEAD_FLAG_RPC)) {
		pr_error("%s: The message is not an RPC\n", __func__);
		err = -EINVAL;
		goto out;
	}



	pin = rt_proto_payload_get((void *) msg);

	/* Run the correspondent action */
	err = rt_action_run(header, pin);
	if (err)
		pr_error("%s: action failure err: %d\n\r", __func__, err);

out:
	mq_discard(0, mq_in_slot);
	return err;
}
