/**
 * @defgroup library-message-queue Message Queue
 * @{
 * @ingroup library
 * @brief Message Queues definitions and functions
 * @copyright (C) 2013-2016 CERN (www.cern.ch)
 * @author Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * @author Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef __TRTL_RT_MQUEUE_H
#define __TRTL_RT_MQUEUE_H

#include <hw/mockturtle_addresses.h>
#include <hw/mockturtle_queue.h>
#include <hw/mockturtle_cpu_lr.h>

/**
 * List of Message Queue types
 */
enum trtl_mq_type {
	TRTL_HMQ = 0, /**< Host Message Queue - Host-Firmware communication */
	TRTL_RMQ, /**< Remote Message Queue - Network-Firmware communication */
};


/**
 * It gets the Message Queue base address
 * @param[in] type MQ type
 */
static inline void *trtl_mq_base_address(enum trtl_mq_type type)
{
	return (void *)(type == TRTL_HMQ ? TRTL_ADDR_HMQ_BASE : TRTL_ADDR_RMQ_BASE);
}


/**
 * It writes on a Message Queue register
 * @param[in] type MQ type to use
 * @param[val] value to write
 * @param[reg] reg register offset
 */
static inline void mq_writel(enum trtl_mq_type type, uint32_t val, uint32_t reg)
{
	*(volatile uint32_t *)(trtl_mq_base_address(type) + reg) = val;
}


/**
 * @copydoc TRTL_MQ_CMD_CLAIM
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_claim(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_CLAIM, TRTL_MQ_BASE_OUT(slot) + TRTL_MQ_SLOT_COMMAND);
}

/**
 * @copydoc TRTL_MQ_CMD_PURGE
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_purge(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_PURGE, TRTL_MQ_BASE_OUT(slot) + TRTL_MQ_SLOT_COMMAND);
}


/**
 * @copydoc TRTL_MQ_CMD_READY
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 * @param[in] count number of 32bit words to send
 */
static inline void mq_send(enum trtl_mq_type type, int slot, int count)
{
	mq_writel(type, TRTL_MQ_CMD_READY | count, TRTL_MQ_BASE_OUT(slot) + TRTL_MQ_SLOT_COMMAND);
}


/**
 * @copydoc TRTL_MQ_CMD_DISCARD
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_discard(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_DISCARD, TRTL_MQ_BASE_IN(slot));
}


/**
 * It gets the output slot data field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_out_buffer(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_BASE_OUT (slot) + TRTL_MQ_SLOT_DATA_START);
}


/**
 * It gets the input slot data field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_in_buffer(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_BASE_IN (slot) + TRTL_MQ_SLOT_DATA_START);
}


/**
 * It gets the output slot header field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_out_header(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_BASE_OUT (slot) + TRTL_MQ_SLOT_HEADER_START);
}


/**
 * It gets the input slot header field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_in_header(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_BASE_IN (slot) + TRTL_MQ_SLOT_HEADER_START);
}

/**
 * It gets the current status of the Message Queues (both Host and Remote)
 */
static inline uint32_t mq_poll(void)
{
	return *(volatile uint32_t *) (TRTL_ADDR_LR_BASE + MT_CPU_LR_REG_POLL);
}


/**
 * It gets the current status of the Remote Message Queues
 */
static inline uint32_t rmq_poll(int slot)
{
	return (mq_poll() & ( 1<< (16 + slot )));
}

#endif
/**@}*/
