/**
 * @defgroup library-shared-memory Shared Memory Utilities
 * @{
 * @ingroup library
 * @brief Shared Memory definitions & API
 * @copyright (C) 2013-2016 CERN (www.cern.ch)
 * @author Federico Vaga <federico.vaga@cern.ch>
 * @author Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef __TRTL_RT_SMEM_H
#define __TRTL_RT_SMEM_H

/**
 * List of supported Shared Memory access type
 */
enum trtl_smem_type {
	TRTL_SMEM_TYPE_BASE = 0, /**< no operation */
	TRTL_SMEM_TYPE_ADD, /**< atomic addition */
	TRTL_SMEM_TYPE_SUB, /**< atomic subtraction */
	TRTL_SMEM_TYPE_SET, /**< atomic bit set */
	TRTL_SMEM_TYPE_CLR, /**< atomic bit clear */
	TRTL_SMEM_TYPE_FLP, /**< atomic bit flip */
	TRTL_SMEM_TYPE_TST_SET, /**< atomic test and set */
};

/**
 * Shared Memory Size
 */
#define TRTL_SMEM_SIZE 0x10000

/**
 * Shared Memory address for the
 * @copydoc TRTL_SMEM_TYPE_BASE
 * memory access
 */
#define TRTL_SMEM_RANGE_BASE 0x00000

/**
 * It generates the SHM address range for a given SHM type
 */
#define TRTL_SMEM_TYPE_TO_RANGE(_type) \
	(TRTL_SMEM_RANGE_BASE + (_type) * TRTL_SMEM_SIZE)

/**
 * Shared Memory base address for the
 * @copydoc TRTL_SMEM_TYPE_ADD
 * memory access
 */
#define TRTL_SMEM_RANGE_ADD \
	TRTL_SMEM_TYPE_TO_RANGE(TRTL_SMEM_TYPE_ADD)

/**
 * Shared Memory base address for the
 * @copydoc TRTL_SMEM_TYPE_SUB
 * memory access
 */
#define TRTL_SMEM_RANGE_SUB \
	TRTL_SMEM_TYPE_TO_RANGE(TRTL_SMEM_TYPE_SUB)

/**
 * Shared Memory address for the
 * @copydoc TRTL_SMEM_TYPE_SET
 * memory access
 */
#define TRTL_SMEM_RANGE_SET \
	TRTL_SMEM_TYPE_TO_RANGE(TRTL_SMEM_TYPE_SET)

/**
 * Shared Memory address for the
 * @copydoc TRTL_SMEM_TYPE_CLR
 * memory access
 */
#define TRTL_SMEM_RANGE_CLEAR \
	TRTL_SMEM_TYPE_TO_RANGE(TRTL_SMEM_TYPE_CLR)

/**
 * Shared Memory address for the
 * @copydoc TRTL_SMEM_TYPE_FLP
 * memory access
 */
#define TRTL_SMEM_RANGE_FLIP \
	TRTL_SMEM_TYPE_TO_RANGE(TRTL_SMEM_TYPE_FLP)

/**
 * Shared Memory address for the
 * @copydoc TRTL_SMEM_TYPE_TST_SET
 * memory access
 */
#define TRTL_SMEM_RANGE_TEST_AND_SET \
	TRTL_SMEM_TYPE_TO_RANGE(TRTL_SMEM_TYPE_TST_SET)

/**
 * Compiler attribute to allocate variable on the Share Memory.
 * Add this to the variable declaration to put it on the Shared Memory.
 *
 *     SMEM int x;
 */
#define SMEM volatile __attribute__((section(".smem")))


/**
 * It performs an operations on the shared memory. What the function
 * performs can be summerized as:
 *
 *     (*p) = (*p) <operation-type> x
 *
 * Use this function only when the operation type that you want to use
 * is not supported yet by the library directly. Otherwise,
 * use the other functions.
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 * @param[in] type operation type
 */
static inline void __smem_atomic_op(volatile int *p, int x,
				    enum trtl_smem_type type)
{
	*(volatile int *)(p + (TRTL_SMEM_TYPE_TO_RANGE(type) >> 2)) = x;
}


/**
 * It performs an
 * @copydoc TRTL_SMEM_TYPE_ADD
 *
 *     (*p) = (*p) + x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_add(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_TYPE_ADD);
}


/**
 * It performs an
 * @copydoc TRTL_SMEM_TYPE_SUB
 *
 *     (*p) = (*p) - x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_sub(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_TYPE_SUB);
}


/**
 * It performs an
 * @copydoc TRTL_SMEM_TYPE_SET
 *
 *     (*p) = (*p) | x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_or(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_TYPE_SET);
}


/**
 * It performs an
 * @copydoc TRTL_SMEM_TYPE_CLR
 *
 *     (*p) = (*p) & (~x)
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_and_not(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_TYPE_CLR);
}


/**
 * It performs an
 * @copydoc TRTL_SMEM_TYPE_FLP
 *
 *     (*p) = (*p) ^ x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_xor(int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_TYPE_FLP);
}


/**
 * It performs an:
 * @copydoc TRTL_SMEM_TYPE_TST_SET
 *
 *     val = (*p);
 *     if (val == 0) {
 *        (*p) = 1;
 *     }
 *     return val;
 *
 * This is useful to implement mutex
 *
 * @param[in] p address on the shared memory
 * @return the value before the set
 */
static inline int smem_atomic_test_and_set(int *p)
{
	/* shift right translates range in bytes into range in words */
	return *(volatile int *)(p + (TRTL_SMEM_TYPE_TST_SET >> 2));
}

#endif
/**@}*/
