/*
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <libmockturtle.h>
#include <getopt.h>
#include <pthread.h>
#include <unistd.h>
#include <time.h>
#include <libgen.h>


#define MAX_DEV 4
#define MAX_SLOT 32
#define MAX_CPU 8


struct trtl_thread_desc {
	struct trtl_dev *trtl;
	uint32_t dev_id;
	int cpu_index[MAX_CPU];
	int n_cpu;
	int slot_index[MAX_SLOT];
	int n_slot;
	int share;
};

static unsigned int  cnt, n;
static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
static int timestamp = 0;
static int protocol;

static void help()
{
	fprintf(stderr, "\n");
	fprintf(stderr, "mockturtle-messages -D 0x<hex-number> -i <number> [options]\n\n");
	fprintf(stderr, "It dumps all messages from a given set of Mock-Turtle slots\n\n");
	fprintf(stderr, "-D   device identificator in hexadecimal format\n");
	fprintf(stderr, "-i   slot index\n");
	fprintf(stderr, "-n   number of total messages to read. The default is 0 (infinite)\n");
	fprintf(stderr, "-t   print message timestamp\n");
	fprintf(stderr, "-s   share all HMQ with other users\n");
	fprintf(stderr, "-p   parse protocol\n");
	fprintf(stderr, "-h   show this help\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
		"You can dump from several devices slots, so the arguments '-D' and '-i' may appear several times. The argument '-i' refers to the previous device id declared\n\n");
	fprintf(stderr,
		"e.g. Dumping messagges from slots 2 and 3 of devices 0x0402 and 0x0382\n\n");
	fprintf(stderr,
		"        mockturtle-messages -D 0x0382 -i 2 -i 3 -D 0x0402 -i 2 -i 3\n\n");
	exit(1);
}


/**
 * It retreives a message from a given slots and it prints its content
 * @param[in] trtl device to use
 * @param[in] hmq slot to read
 */
static int dump_message(struct trtl_dev *trtl, struct trtl_hmq *hmq)
{
	struct trtl_msg *wmsg;
	time_t tm;
	char stime[64];
	struct tm *gm;
	int i;

	if (timestamp) {
		tm = time(NULL);
		gm = gmtime(&tm);
		strftime(stime, 64,"%T", gm);
		fprintf(stdout, "[%s] ", stime);
	}
	fprintf(stdout, "%s :", basename(hmq->syspath));
	wmsg = trtl_hmq_receive(hmq);
	if (!wmsg) {
		fprintf(stdout, " error : %s\n", trtl_strerror(errno));
		return -1;
	}

	fprintf(stdout, "\n");
	if (protocol) {
		trtl_print_message(wmsg);
	} else {
		for (i = 0; i < TRTL_MAX_PAYLOAD_SIZE; ++i) {
			if (i % 4 == 0)
				fprintf(stdout, "\n    %04d :", i);
			fprintf(stdout, " 0x%08x", wmsg->data[i]);
		}

	}
	fprintf(stdout, "\n");

	free(wmsg);

	return 0;
}


/**
 * pthread for each device. It dumps messages from slots
 * @param[in] arg a pointer to the device index
 */
void *message_thread(void *arg)
{
	struct trtl_thread_desc *th_data = arg;
	struct pollfd p[MAX_SLOT];
	struct trtl_dev *trtl = th_data->trtl;
	struct trtl_hmq *hmq[th_data->n_slot];
	int ret, err, i;

	if (!th_data->n_slot)
		return NULL;

	/* Open the device */
	if (!trtl)
		trtl = trtl_open_by_id(th_data->dev_id);

	if (!trtl) {
		fprintf(stderr, "Cannot open Mock Turtle device: %s\n", trtl_strerror(errno));
		pthread_exit(NULL);
	}

	/* Build the polling structures */
	for (i = 0; i < th_data->n_slot; ++i) {
		trtl_hmq_share_set(trtl, 0 , th_data->slot_index[i], 1);

		hmq[i] = trtl_hmq_open(trtl, th_data->slot_index[i],
				    TRTL_HMQ_OUTCOMING);
		if (!hmq[i]) {
			fprintf(stderr, "Cannot open HMQ: %s\n",
				trtl_strerror(errno));
			goto out;
		}
		p[i].fd = hmq[i]->fd;
		p[i].events = POLLIN | POLLERR;
	}

	/* Start dumping messages */
	while (n == 0 || n > cnt) {
		/* Polling slots */
		ret = poll(p, th_data->n_slot, 10);
		switch (ret) {
		default:
			/* Dump from the slot */
			for (i = 0; i < th_data->n_slot; ++i) {
				if (!(p[i].revents & POLLIN))
					continue;

				err = dump_message(trtl, hmq[i]);
				if (err)
					continue;
				pthread_mutex_lock(&mtx);
				cnt++;
				pthread_mutex_unlock(&mtx);
			}
			break;
		case 0:
			/* timeout */
			break;
		case -1:
			/* error */
			goto out;
		}
	}

out:
	/* Close all message slots */
	while (--i >= 0)
		trtl_hmq_close(hmq[i]);
	trtl_close(trtl);
	return NULL;
}


int main(int argc, char *argv[])
{
	struct trtl_thread_desc th_data[MAX_DEV], *last;
	unsigned long i;
	unsigned int di = 0;
	int share = 0;
	pthread_t tid_msg[MAX_DEV];
	int err;
	char c;

	atexit(trtl_exit);

	memset(th_data, 0, sizeof(struct trtl_thread_desc) * MAX_DEV);

	while ((c = getopt (argc, argv, "hi:D:n:tsp")) != -1) {
		switch (c) {
		default:
			help();
			break;
		case 's':
			share = 1;
			break;
		case 'i':
		/* Save slot index for each device id */
			if (!last || last->n_slot >= MAX_SLOT)
				break;
			sscanf(optarg, "%d",
			       &last->slot_index[last->n_slot]);
			last->n_slot++;
			break;
		case 'D':
		/* Save device ids to use */
			if (di >= MAX_DEV)
				break;
			last = &th_data[di];
			sscanf(optarg, "0x%x", &last->dev_id);
			di++;
			break;
		case 'n':
		/* Number of total messages to dump */
			sscanf(optarg, "%d", &n);
			break;
		case 't':
			timestamp = 1;
			break;
		case 'p':
			protocol = 1;
			break;
		}
	}

	if (argc == 1) {
		help();
		exit(1);
	}

	err = trtl_init();
	if (err) {
		fprintf(stderr, "Cannot init White Rabbit Node Core lib: %s\n",
			trtl_strerror(errno));
		exit(1);
	}

	/* Set global sharing option */
	for (i = 0; i < di; ++i)
		th_data[i].share = share;

	/* Run dumping on in parallel from several devices */
	for (i = 0; i < di; i++) {
		err = pthread_create(&tid_msg[i], NULL, message_thread,
				     (void *)(&th_data[i]));
		if (err)
			fprintf(stderr,
				"Cannot create 'dump_thread' instance %ld: %s\n",
				i, strerror(errno));
	}


	/* Wait for the threads to finish */
	for (i = 0; i < di; i++)
		pthread_join(tid_msg[i], NULL);
	exit(0);
}
