/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <inttypes.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <libmockturtle.h>


/**
 * It prints the help message
 * @param[in] name the program name
 */
static void help(char *name)
{
	fprintf(stderr, "%s [options]\n", name);
	fprintf(stderr, "  -h             print this help\n");
	fprintf(stderr, "  -D 0x<dev_id>  device id to ping\n");
	fprintf(stderr, "  -i <num>       input slot number\n");
	fprintf(stderr, "  -o <num>       output slot number\n");
	fprintf(stderr, "  -n <num>       number of ping to perform\n");
	fprintf(stderr, "  -p <num>       ping period in micro-seconds\n");
	fprintf(stderr, "  -v             show device version\n");
}


/**
 * It prints the given Mock Turtle firmware version
 * @param[in] v version to print
 */
static void trtl_ping_version_print(struct trtl_rt_version *v)
{
	fprintf(stdout, "\tFirmware Version:\n");
	fprintf(stdout, "\t    bitstream_id: 0x%08X\n",
		v->fpga_id);
	fprintf(stdout, "\t  application_id: 0x%08X\n",
		v->rt_id);
	fprintf(stdout, "\tfirmware_version: 0x%08X\n",
		v->rt_version);
	fprintf(stdout, "\t     git_version: 0x%08X\n",
		v->git_version);
}


int main(int argc, char *argv[])
{
	struct trtl_dev *trtl;
	uint32_t dev_id = 0, n = 1;
	struct trtl_rt_version version;
	uint64_t period = 0;
	int err, f_version = 0, hmq_i = 0, hmq_o = 0;
	char c;

	while ((c = getopt (argc, argv, "hD:n:p:tvi:o:")) != -1) {
		switch (c) {
		case 'h':
		case '?':
			help(argv[0]);
			exit(1);
			break;
		case 'D':
			sscanf(optarg, "0x%x", &dev_id);
			break;
		case 'n':
			sscanf(optarg, "%d", &n);
			break;
		case 'p':
			sscanf(optarg, "%"SCNu64, &period);
			break;
		case 'v':
			f_version = 1;
			break;
		case 'i':
			sscanf(optarg, "%d", &hmq_i);
			break;
		case 'o':
			sscanf(optarg, "%d", &hmq_o);
			break;
		}
	}


	atexit(trtl_exit);
	err = trtl_init();
	if (err) {
		fprintf(stderr,
			"Cannot init Mock Turtle lib: %s\n",
			trtl_strerror(errno));
		exit(1);
	}

	trtl = trtl_open_by_id(dev_id);
	if (!trtl) {
		fprintf(stderr,
			"Cannot open Mock Turtle device 0x%x: %s\n",
			dev_id, trtl_strerror(errno));
		exit(1);
	}

	while (n--) {
		fprintf(stdout, "\n");
		usleep(period);

		/* Try to ping */
		err = trtl_rt_ping(trtl, hmq_i, hmq_o);
		if (err) {
			fprintf(stderr,
				"Ping failed on Device 0x%x HMQ [in:%d, out:%d]: %s\n",
				dev_id, hmq_i, hmq_o, trtl_strerror(errno));
			continue;
		}

		fprintf(stdout,
			"Something running and connected to [device: 0x%x, in:%d, out:%d]!\n",
			dev_id, hmq_i, hmq_o);

		/* Get the version */
		if (f_version) {
			trtl_rt_version_get(trtl, &version, hmq_i, hmq_o);
			trtl_ping_version_print(&version);
		}

	}

	trtl_close(trtl);
	exit(0);
}
