/*
 * Copyright (C) 2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <libmockturtle.h>

static void help(char *name)
{
	fprintf(stderr, "\n");
	fprintf(stderr,
		"%s -D 0x<hex-number> -i <number> -o <number> [write <idx> <value> <idx> <value> ...] [read <idx> <idx> ...]\n\n",
		name);
	fprintf(stderr, "It reads/writes variables on a HMQ\n\n");
	fprintf(stderr, "-D   device identificator in hexadecimal format\n");
	fprintf(stderr, "-i   slot index where send\n");
	fprintf(stderr, "-o   slot index where receive\n");
	fprintf(stderr, "-h   show this help\n");
	fprintf(stderr, "\n");
	exit(1);
}

enum operations {
	OP_READ,
	OP_WRITE,
};


int main(int argc, char *argv[])
{
	struct trtl_proto_header hdr;
	struct trtl_dev *trtl;
	char c;
	unsigned int n_var;
	int err, n, dev_id = -1, val, i, v;
	uint32_t *var;
	enum operations mode;

	memset(&hdr, 0, sizeof(struct trtl_proto_header));
	atexit(trtl_exit);

	while ((c = getopt (argc, argv, "hD:i:o:")) != -1) {
		switch (c) {
		default:
			help(argv[0]);
			break;
		case 'D':
			n = sscanf(optarg, "0x%x", &dev_id);
			if (n != 1) {
				fprintf(stderr, "Invalid Device ID\n");
				exit(1);
			}
			break;
		case 'i':
			n = sscanf(optarg, "%d", &val);
			if (n != 1) {
				fprintf(stderr, "Invalid HMQ number\n");
				exit(1);
			}
			hdr.slot_io = (hdr.slot_io & ~0xF) | (val & 0xF);
			break;
		case 'o':
			n = sscanf(optarg, "%d", &val);
			if (n != 1) {
				fprintf(stderr, "Invalid HMQ number\n");
				exit(1);
			}
			val =  val << 4;
			hdr.slot_io = (hdr.slot_io & ~0xF0) | (val & 0xF0);
			break;
		}
	}
	if (strcmp(argv[optind], "write") == 0) {
		mode = OP_WRITE;
	} else if (strcmp(argv[optind], "read") == 0) {
		mode = OP_READ;
	} else {
		fprintf(stderr, "Unknown operation \'%s\' (read, write)\n",
			argv[optind]);
		exit(1);
	}

	optind++;
	n_var = argc - optind;
	if (n_var <= 0 || (strcmp(argv[optind], "write") == 0 && n_var % 2)) {
		fprintf(stderr, "Invalid number of arguments\n");
		exit(1);
	}
	if (mode == OP_WRITE)
		n_var /= 2;

	/* Build variable message payload */
	var = calloc(n_var * 2, sizeof(uint32_t));
	if (!var)
		exit(1);

	for (i = optind, v = 0; i < argc;
	     i += (mode == OP_READ ? 1 : 2), v += 2) {
		n = sscanf(argv[i], "%d", &var[v]);
		if (n != 1) {
			fprintf(stderr,
				"Invalid index: it must be a decimal number (%s)\n",
				argv[i]);
			exit(1);
		}
		if (mode == OP_READ)
			continue;

		/* Write mode only */
		n = sscanf(argv[i + 1], "0x%x", &var[v + 1]);
		if (n == 1)
			continue;
		n = sscanf(argv[i + 1], "%d", &var[v + 1]);
		if (n != 1) {
			fprintf(stderr,
				"Invalid value: it must be a decimal or hexadecimal number (%s)\n",
				argv[i + 1]);
			exit(1);
		}
	}


	err = trtl_init();
	if (err) {
		fprintf(stderr, "Cannot init White Rabbit Node Core lib: %s\n",
			trtl_strerror(errno));
		exit(1);
	}

	trtl = trtl_open_by_id(dev_id);
	if (!trtl) {
		fprintf(stderr, "Cannot open TRTL: %s\n", trtl_strerror(errno));
		exit(errno);
	}

	switch (mode) {
	case OP_READ:
		err = trtl_rt_variable_get(trtl, &hdr, var, n_var);
		if (err) {
			fprintf(stderr, "Cannot read variables: %s\n",
				trtl_strerror(errno));
			break;
		}
		for (i = 0; i < n_var; ++i) {
			fprintf(stdout, "[%d] 0x%x\n",
				var[i * 2], var[i * 2 + 1]);
		}
		break;
	case OP_WRITE:
		err = trtl_rt_variable_set(trtl, &hdr, var, n_var);
		if (err) {
			fprintf(stderr, "Cannot write variables: %s\n",
				trtl_strerror(errno));
		}
		break;
	}

	trtl_close(trtl);
	if (err) {
		exit(errno);
	} else {
		exit(0);
	}
}
